# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Traits extension
................

This module contains Trait classes that we've pulled from the
traits source and fixed due to various bugs.  File and Directory are
redefined as the release version had dependencies on TraitsUI, which
we do not want Nipype to depend on.  At least not yet.

Undefined class was missing the __len__ operator, causing edit_traits
and configure_traits to fail on List objects.  Even though we don't
require TraitsUI, this bug was the only thing preventing us from
popping up GUIs which users like.

These bugs have been in Traits v3.3.0 and v3.2.1.  We have reported
all of these bugs and they've been fixed in enthought svn repository
(usually by Robert Kern).

"""
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import str, bytes
import os
import collections

# perform all external trait imports here
from traits import __version__ as traits_version
import traits.api as traits
from traits.trait_handlers import TraitDictObject, TraitListObject
from traits.trait_errors import TraitError
from traits.trait_base import _Undefined, class_of

from traits.api import BaseUnicode
from traits.api import Unicode
from future import standard_library

if traits_version < '3.7.0':
    raise ImportError('Traits version 3.7.0 or higher must be installed')

standard_library.install_aliases()


class Str(Unicode):
    """Replacement for the default traits.Str based in bytes"""


# Monkeypatch Str and DictStrStr for Python 2 compatibility
traits.Str = Str
DictStrStr = traits.Dict((bytes, str), (bytes, str))
traits.DictStrStr = DictStrStr


class File(BaseUnicode):
    """ Defines a trait whose value must be the name of a file.
    """

    # A description of the type of value this trait accepts:
    info_text = 'a file name'

    def __init__(self,
                 value='',
                 filter=None,
                 auto_set=False,
                 entries=0,
                 exists=False,
                 **metadata):
        """ Creates a File trait.

        Parameters
        ----------
        value : string
            The default value for the trait
        filter : string
            A wildcard string to filter filenames in the file dialog box used by
            the attribute trait editor.
        auto_set : boolean
            Indicates whether the file editor updates the trait value after
            every key stroke.
        exists : boolean
            Indicates whether the trait value must be an existing file or
            not.

        Default Value
        -------------
        *value* or ''
        """
        self.filter = filter
        self.auto_set = auto_set
        self.entries = entries
        self.exists = exists

        if exists:
            self.info_text = 'an existing file name'

        super(File, self).__init__(value, **metadata)

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait."""
        validated_value = super(File, self).validate(object, name, value)
        if not self.exists:
            return validated_value
        elif os.path.isfile(value):
            return validated_value
        else:
            raise TraitError(
                args='The trait \'{}\' of {} instance is {}, but the path '
                ' \'{}\' does not exist.'.format(name, class_of(object),
                                                 self.info_text, value))

        self.error(object, name, value)


# -------------------------------------------------------------------------------
#  'Directory' trait
# -------------------------------------------------------------------------------


class Directory(BaseUnicode):
    """
    Defines a trait whose value must be the name of a directory.
    """

    # A description of the type of value this trait accepts:
    info_text = 'a directory name'

    def __init__(self,
                 value='',
                 auto_set=False,
                 entries=0,
                 exists=False,
                 **metadata):
        """ Creates a Directory trait.

        Parameters
        ----------
        value : string
            The default value for the trait
        auto_set : boolean
            Indicates whether the directory editor updates the trait value
            after every key stroke.
        exists : boolean
            Indicates whether the trait value must be an existing directory or
            not.

        Default Value
        -------------
        *value* or ''
        """
        self.entries = entries
        self.auto_set = auto_set
        self.exists = exists

        if exists:
            self.info_text = 'an existing directory name'

        super(Directory, self).__init__(value, **metadata)

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait."""
        if isinstance(value, (str, bytes)):
            if not self.exists:
                return value
            if os.path.isdir(value):
                return value
            else:
                raise TraitError(
                    args='The trait \'{}\' of {} instance is {}, but the path '
                    ' \'{}\' does not exist.'.format(name, class_of(object),
                                                     self.info_text, value))

        self.error(object, name, value)


# lists of tuples
# each element consists of :
# - uncompressed (tuple[0]) extension
# - compressed (tuple[1]) extension
img_fmt_types = {
    'nifti1': [('.nii', '.nii.gz'), (('.hdr', '.img'), ('.hdr', '.img.gz'))],
    'mgh': [('.mgh', '.mgz'), ('.mgh', '.mgh.gz')],
    'nifti2': [('.nii', '.nii.gz')],
    'cifti2': [('.nii', '.nii.gz')],
    'gifti': [('.gii', '.gii.gz')],
    'dicom': [('.dcm', '.dcm'), ('.IMA', '.IMA'), ('.tar', '.tar.gz')],
    'nrrd': [('.nrrd', 'nrrd'), ('nhdr', 'nhdr')],
    'afni': [('.HEAD', '.HEAD'), ('.BRIK', '.BRIK')]
}


class ImageFile(File):
    """ Defines a trait of specific neuroimaging files """

    def __init__(self,
                 value='',
                 filter=None,
                 auto_set=False,
                 entries=0,
                 exists=False,
                 types=[],
                 allow_compressed=True,
                 **metadata):
        """ Trait handles neuroimaging files.

        Parameters
        ----------
        types : list
            Strings of file format types accepted
        compressed : boolean
            Indicates whether the file format can compressed
        """
        self.types = types
        self.allow_compressed = allow_compressed
        super(ImageFile, self).__init__(value, filter, auto_set, entries,
                                        exists, **metadata)

    def info(self):
        existing = 'n existing' if self.exists else ''
        comma = ',' if self.exists and not self.allow_compressed else ''
        uncompressed = ' uncompressed' if not self.allow_compressed else ''
        with_ext = ' (valid extensions: [{}])'.format(
            ', '.join(self.grab_exts())) if self.types else ''
        return 'a{existing}{comma}{uncompressed} file{with_ext}'.format(
            existing=existing, comma=comma, uncompressed=uncompressed,
            with_ext=with_ext)

    def grab_exts(self):
        # TODO: file type validation
        exts = []
        for fmt in self.types:
            if fmt in img_fmt_types:
                exts.extend(
                    sum([[u for u in y[0]]
                         if isinstance(y[0], tuple) else [y[0]]
                         for y in img_fmt_types[fmt]], []))
                if self.allow_compressed:
                    exts.extend(
                        sum([[u for u in y[-1]]
                             if isinstance(y[-1], tuple) else [y[-1]]
                             for y in img_fmt_types[fmt]], []))
            else:
                raise AttributeError(
                    'Information has not been added for format'
                    ' type {} yet. Supported formats include: '
                    '{}'.format(fmt, ', '.join(img_fmt_types.keys())))
        return list(set(exts))

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait.
        """
        validated_value = super(ImageFile, self).validate(object, name, value)
        if validated_value and self.types:
            _exts = self.grab_exts()
            if not any(validated_value.endswith(x) for x in _exts):
                raise TraitError(
                    args="{} is not included in allowed types: {}".format(
                        validated_value, ', '.join(_exts)))
        return validated_value


"""
The functions that pop-up the Traits GUIs, edit_traits and
configure_traits, were failing because all of our inputs default to
Undefined deep and down in traits/ui/wx/list_editor.py it checks for
the len() of the elements of the list.  The _Undefined class in traits
does not define the __len__ method and would error.  I tried defining
our own Undefined and even sublassing Undefined, but both of those
failed with a TraitError in our initializer when we assign the
Undefined to the inputs because of an incompatible type:

TraitError: The 'vertical_gradient' trait of a BetInputSpec instance must be a float, but a value of <undefined> <class 'nipype.interfaces.traits._Undefined'> was specified.

So... in order to keep the same type but add the missing method, I
monkey patched.
"""


def length(self):
    return 0


##########################################################################
# Apply monkeypatch here
_Undefined.__len__ = length
##########################################################################

Undefined = _Undefined()


def isdefined(object):
    return not isinstance(object, _Undefined)


def has_metadata(trait, metadata, value=None, recursive=True):
    '''
    Checks if a given trait has a metadata (and optionally if it is set to particular value)
    '''
    count = 0
    if hasattr(trait, "_metadata") and metadata in list(
            trait._metadata.keys()) and (trait._metadata[metadata] == value
                                         or value is None):
        count += 1
    if recursive:
        if hasattr(trait, 'inner_traits'):
            for inner_trait in trait.inner_traits():
                count += has_metadata(inner_trait.trait_type, metadata,
                                      recursive)
        if hasattr(trait, 'handlers') and trait.handlers is not None:
            for handler in trait.handlers:
                count += has_metadata(handler, metadata, recursive)

    return count > 0


class MultiObject(traits.List):
    """ Abstract class - shared functionality of input and output MultiObject
    """

    def validate(self, object, name, value):

        # want to treat range and other sequences (except str) as list
        if not isinstance(value, (str, bytes)) and isinstance(
                value, collections.Sequence):
            value = list(value)

        if not isdefined(value) or \
                (isinstance(value, list) and len(value) == 0):
            return Undefined

        newvalue = value

        inner_trait = self.inner_traits()[0]
        if not isinstance(value, list) \
            or (isinstance(inner_trait.trait_type, traits.List) and
                not isinstance(inner_trait.trait_type, InputMultiObject) and
                not isinstance(value[0], list)):
            newvalue = [value]
        value = super(MultiObject, self).validate(object, name, newvalue)

        if value:
            return value

        self.error(object, name, value)


class OutputMultiObject(MultiObject):
    """ Implements a user friendly traits that accepts one or more
    paths to files or directories. This is the output version which
    return a single string whenever possible (when it was set to a
    single value or a list of length 1). Default value of this trait
    is _Undefined. It does not accept empty lists.

    XXX This should only be used as a final resort. We should stick to
    established Traits to the extent possible.

    XXX This needs to be vetted by somebody who understands traits

    >>> from nipype.interfaces.base import OutputMultiObject, TraitedSpec
    >>> class A(TraitedSpec):
    ...     foo = OutputMultiObject(File(exists=False))
    >>> a = A()
    >>> a.foo
    <undefined>

    >>> a.foo = '/software/temp/foo.txt'
    >>> a.foo
    '/software/temp/foo.txt'

    >>> a.foo = ['/software/temp/foo.txt']
    >>> a.foo
    '/software/temp/foo.txt'

    >>> a.foo = ['/software/temp/foo.txt', '/software/temp/goo.txt']
    >>> a.foo
    ['/software/temp/foo.txt', '/software/temp/goo.txt']

    """

    def get(self, object, name):
        value = self.get_value(object, name)
        if len(value) == 0:
            return Undefined
        elif len(value) == 1:
            return value[0]
        else:
            return value

    def set(self, object, name, value):
        self.set_value(object, name, value)


class InputMultiObject(MultiObject):
    """ Implements a user friendly traits that accepts one or more
    paths to files or directories. This is the input version which
    always returns a list. Default value of this trait
    is _Undefined. It does not accept empty lists.

    XXX This should only be used as a final resort. We should stick to
    established Traits to the extent possible.

    XXX This needs to be vetted by somebody who understands traits

    >>> from nipype.interfaces.base import InputMultiObject, TraitedSpec
    >>> class A(TraitedSpec):
    ...     foo = InputMultiObject(File(exists=False))
    >>> a = A()
    >>> a.foo
    <undefined>

    >>> a.foo = '/software/temp/foo.txt'
    >>> a.foo
    ['/software/temp/foo.txt']

    >>> a.foo = ['/software/temp/foo.txt']
    >>> a.foo
    ['/software/temp/foo.txt']

    >>> a.foo = ['/software/temp/foo.txt', '/software/temp/goo.txt']
    >>> a.foo
    ['/software/temp/foo.txt', '/software/temp/goo.txt']

    """
    pass

InputMultiPath = InputMultiObject
OutputMultiPath = OutputMultiObject



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest
import os
import mock
import numpy as np

from .....testing import utils
from .....interfaces import IdentityInterface
from .....pipeline.engine import Node, Workflow

from ..resting import create_resting_preproc

ALL_FIELDS = [
    'func', 'in_file', 'slice_time_corrected_file', 'stddev_file', 'out_stat',
    'thresh', 'num_noise_components', 'detrended_file', 'design_file',
    'highpass_sigma', 'lowpass_sigma', 'out_file', 'noise_mask_file',
    'filtered_file'
]


def stub_node_factory(*args, **kwargs):
    if 'name' not in kwargs.keys():
        raise Exception()
    name = kwargs['name']
    if name == 'compcor':
        return Node(*args, **kwargs)
    else:  # replace with an IdentityInterface
        return Node(IdentityInterface(fields=ALL_FIELDS), name=name)


def stub_wf(*args, **kwargs):
    wflow = Workflow(name='realigner')
    inputnode = Node(IdentityInterface(fields=['func']), name='inputspec')
    outputnode = Node(
        interface=IdentityInterface(fields=['realigned_file']),
        name='outputspec')
    wflow.connect(inputnode, 'func', outputnode, 'realigned_file')
    return wflow


class TestResting():

    in_filenames = {
        'realigned_file': 'rsfmrifunc.nii',
        'mask_file': 'rsfmrimask.nii'
    }

    out_filenames = {
        'components_file': 'restpreproc/compcor/noise_components.txt'
    }

    num_noise_components = 6

    @pytest.fixture(autouse=True)
    def setup_class(self, tmpdir):
        # setup temp folder
        tmpdir.chdir()
        self.in_filenames = {
            key: os.path.abspath(value)
            for key, value in self.in_filenames.items()
        }

        # create&save input files
        utils.save_toy_nii(self.fake_data, self.in_filenames['realigned_file'])
        mask = np.zeros(self.fake_data.shape[:3])
        for i in range(mask.shape[0]):
            for j in range(mask.shape[1]):
                if i == j:
                    mask[i, j] = 1
        utils.save_toy_nii(mask, self.in_filenames['mask_file'])

    @mock.patch(
        'nipype.workflows.rsfmri.fsl.resting.create_realign_flow',
        side_effect=stub_wf)
    @mock.patch('nipype.pipeline.engine.Node', side_effect=stub_node_factory)
    def test_create_resting_preproc(self, mock_node, mock_realign_wf):
        wflow = create_resting_preproc(base_dir=os.getcwd())

        wflow.inputs.inputspec.num_noise_components = self.num_noise_components
        mask_in = wflow.get_node('threshold').inputs
        mask_in.out_file = self.in_filenames['mask_file']
        func_in = wflow.get_node('slicetimer').inputs
        func_in.slice_time_corrected_file = self.in_filenames['realigned_file']

        wflow.run()

        # assert
        expected_file = os.path.abspath(self.out_filenames['components_file'])
        with open(expected_file, 'r') as components_file:
            components_data = [line.split() for line in components_file]
            num_got_components = len(components_data)
            assert (num_got_components == self.num_noise_components
                    or num_got_components == self.fake_data.shape[3])
            first_two = [row[:2] for row in components_data[1:]]
            assert first_two == [['-0.5172356654', '-0.6973053243'], [
                '0.2574722644', '0.1645270737'
            ], ['-0.0806469590',
                '0.5156853779'], ['0.7187176051', '-0.3235820287'],
                                 ['-0.3783072450', '0.3406749013']]

    fake_data = np.array([[[[2, 4, 3, 9, 1], [3, 6, 4, 7, 4]],
                           [[8, 3, 4, 6, 2], [4, 0, 4, 4, 2]]],
                          [[[9, 7, 5, 5, 7], [7, 8, 4, 8, 4]],
                           [[0, 4, 7, 1, 7], [6, 8, 8, 8, 7]]]])



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from __future__ import unicode_literals
from ..preprocess import Automask


def test_Automask_inputs():
    input_map = dict(
        args=dict(argstr='%s', ),
        brain_file=dict(
            argstr='-apply_prefix %s',
            name_source='in_file',
            name_template='%s_masked',
        ),
        clfrac=dict(argstr='-clfrac %s', ),
        dilate=dict(argstr='-dilate %s', ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        erode=dict(argstr='-erode %s', ),
        ignore_exception=dict(
            deprecated='1.0.0',
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr='%s',
            copyfile=False,
            mandatory=True,
            position=-1,
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr='-prefix %s',
            name_source='in_file',
            name_template='%s_mask',
        ),
        outputtype=dict(),
        terminal_output=dict(
            deprecated='1.0.0',
            nohash=True,
        ),
    )
    inputs = Automask.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value
def test_Automask_outputs():
    output_map = dict(
        brain_file=dict(),
        out_file=dict(),
    )
    outputs = Automask.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from builtins import str, open, bytes
# This tool exports a Nipype interface in the Boutiques (https://github.com/boutiques) JSON format.
# Boutiques tools can be imported in CBRAIN (https://github.com/aces/cbrain) among other platforms.
#
# Limitations:
#   * List outputs are not supported.
#   * Default values are not extracted from the documentation of the Nipype interface.
#   * The following input types must be ignored for the output path template creation (see option -t):
#     ** String restrictions, i.e. String inputs that accept only a restricted set of values.
#     ** mutually exclusive inputs.
#   * Path-templates are wrong when output files are not created in the execution directory (e.g. when a sub-directory is created).
#   * Optional outputs, i.e. outputs that not always produced, may not be detected.

import os
import argparse
import sys
import tempfile
import simplejson as json

from ..scripts.instance import import_module


def generate_boutiques_descriptor(
        module, interface_name, ignored_template_inputs, docker_image,
        docker_index, verbose, ignore_template_numbers):
    '''
    Returns a JSON string containing a JSON Boutiques description of a Nipype interface.
    Arguments:
    * module: module where the Nipype interface is declared.
    * interface: Nipype interface.
    * ignored_template_inputs: a list of input names that should be ignored in the generation of output path templates.
    * ignore_template_numbers: True if numbers must be ignored in output path creations.
    '''

    if not module:
        raise Exception("Undefined module.")

    # Retrieves Nipype interface
    if isinstance(module, (str, bytes)):
        import_module(module)
        module_name = str(module)
        module = sys.modules[module]
    else:
        module_name = str(module.__name__)

    interface = getattr(module, interface_name)()
    inputs = interface.input_spec()
    outputs = interface.output_spec()

    # Tool description
    tool_desc = {}
    tool_desc['name'] = interface_name
    tool_desc[
        'command-line'] = "nipype_cmd " + module_name + " " + interface_name + " "
    tool_desc[
        'description'] = interface_name + ", as implemented in Nipype (module: " + module_name + ", interface: " + interface_name + ")."
    tool_desc['inputs'] = []
    tool_desc['outputs'] = []
    tool_desc['tool-version'] = interface.version
    tool_desc['schema-version'] = '0.2-snapshot'
    if docker_image:
        tool_desc['docker-image'] = docker_image
    if docker_index:
        tool_desc['docker-index'] = docker_index

    # Generates tool inputs
    for name, spec in sorted(interface.inputs.traits(transient=None).items()):
        input = get_boutiques_input(inputs, interface, name, spec,
                                    ignored_template_inputs, verbose,
                                    ignore_template_numbers)
        tool_desc['inputs'].append(input)
        tool_desc['command-line'] += input['command-line-key'] + " "
        if verbose:
            print("-> Adding input " + input['name'])

    # Generates tool outputs
    for name, spec in sorted(outputs.traits(transient=None).items()):
        output = get_boutiques_output(name, interface, tool_desc['inputs'],
                                      verbose)
        if output['path-template'] != "":
            tool_desc['outputs'].append(output)
            if verbose:
                print("-> Adding output " + output['name'])
        elif verbose:
            print("xx Skipping output " + output['name'] +
                  " with no path template.")
    if tool_desc['outputs'] == []:
        raise Exception("Tool has no output.")

    # Removes all temporary values from inputs (otherwise they will
    # appear in the JSON output)
    for input in tool_desc['inputs']:
        del input['tempvalue']

    return json.dumps(tool_desc, indent=4, separators=(',', ': '))


def get_boutiques_input(inputs, interface, input_name, spec,
                        ignored_template_inputs, verbose,
                        ignore_template_numbers):
    """
    Returns a dictionary containing the Boutiques input corresponding to a Nipype intput.

    Args:
      * inputs: inputs of the Nipype interface.
      * interface: Nipype interface.
      * input_name: name of the Nipype input.
      * spec: Nipype input spec.
      * ignored_template_inputs: input names for which no temporary value must be generated.
      * ignore_template_numbers: True if numbers must be ignored in output path creations.

    Assumes that:
      * Input names are unique.
    """
    if not spec.desc:
        spec.desc = "No description provided."
    spec_info = spec.full_info(inputs, input_name, None)

    input = {}
    input['id'] = input_name
    input['name'] = input_name.replace('_', ' ').capitalize()
    input['type'] = get_type_from_spec_info(spec_info)
    input['list'] = is_list(spec_info)
    input['command-line-key'] = "[" + input_name.upper(
    ) + "]"  # assumes that input names are unique
    input['command-line-flag'] = ("--%s" % input_name + " ").strip()
    input['tempvalue'] = None
    input['description'] = spec_info.capitalize(
    ) + ". " + spec.desc.capitalize()
    if not input['description'].endswith('.'):
        input['description'] += '.'
    if not (hasattr(spec, "mandatory") and spec.mandatory):
        input['optional'] = True
    else:
        input['optional'] = False
    if spec.usedefault:
        input['default-value'] = spec.default_value()[1]

    # Create unique, temporary value.
    temp_value = must_generate_value(input_name, input['type'],
                                     ignored_template_inputs, spec_info, spec,
                                     ignore_template_numbers)
    if temp_value:
        tempvalue = get_unique_value(input['type'], input_name)
        setattr(interface.inputs, input_name, tempvalue)
        input['tempvalue'] = tempvalue
        if verbose:
            print("oo Path-template creation using " + input['id'] + "=" +
                  str(tempvalue))

    # Now that temp values have been generated, set Boolean types to
    # Number (there is no Boolean type in Boutiques)
    if input['type'] == "Boolean":
        input['type'] = "Number"

    return input


def get_boutiques_output(name, interface, tool_inputs, verbose=False):
    """
    Returns a dictionary containing the Boutiques output corresponding to a Nipype output.

    Args:
      * name: name of the Nipype output.
      * interface: Nipype interface.
      * tool_inputs: list of tool inputs (as produced by method get_boutiques_input).

    Assumes that:
      * Output names are unique.
      * Input values involved in the path template are defined.
      * Output files are written in the current directory.
      * There is a single output value (output lists are not supported).
    """
    output = {}
    output['name'] = name.replace('_', ' ').capitalize()
    output['id'] = name
    output['type'] = "File"
    output['path-template'] = ""
    output[
        'optional'] = True  # no real way to determine if an output is always produced, regardless of the input values.

    # Path template creation.

    output_value = interface._list_outputs()[name]
    if output_value != "" and isinstance(
            output_value,
            str):  # FIXME: this crashes when there are multiple output values.
        # Go find from which input value it was built
        for input in tool_inputs:
            if not input['tempvalue']:
                continue
            input_value = input['tempvalue']
            if input['type'] == "File":
                # Take the base name
                input_value = os.path.splitext(
                    os.path.basename(input_value))[0]
            if str(input_value) in output_value:
                output_value = os.path.basename(
                    output_value.replace(input_value,
                                         input['command-line-key'])
                )  # FIXME: this only works if output is written in the current directory
        output['path-template'] = os.path.basename(output_value)
    return output


def get_type_from_spec_info(spec_info):
    '''
    Returns an input type from the spec info. There must be a better
    way to get an input type in Nipype than to parse the spec info.
    '''
    if ("an existing file name" in spec_info) or (
            "input volumes" in spec_info):
        return "File"
    elif ("an integer" in spec_info or "a float" in spec_info):
        return "Number"
    elif "a boolean" in spec_info:
        return "Boolean"
    return "String"


def is_list(spec_info):
    '''
    Returns True if the spec info looks like it describes a list
    parameter. There must be a better way in Nipype to check if an input
    is a list.
    '''
    if "a list" in spec_info:
        return True
    return False


def get_unique_value(type, id):
    '''
    Returns a unique value of type 'type', for input with id 'id',
    assuming id is unique.
    '''
    return {
        "File": os.path.abspath(create_tempfile()),
        "Boolean": True,
        "Number": abs(hash(id)),  # abs in case input param must be positive...
        "String": id
    }[type]


def create_tempfile():
    '''
    Creates a temp file and returns its name.
    '''
    fileTemp = tempfile.NamedTemporaryFile(delete=False)
    fileTemp.write(b"hello")
    fileTemp.close()
    return fileTemp.name


def must_generate_value(name, type, ignored_template_inputs, spec_info, spec,
                        ignore_template_numbers):
    '''
    Return True if a temporary value must be generated for this input.
    Arguments:
    * name: input name.
    * type: input_type.
    * ignored_template_inputs: a list of inputs names for which no value must be generated.
    * spec_info: spec info of the Nipype input
    * ignore_template_numbers: True if numbers must be ignored.
    '''
    # Return false when type is number and numbers must be ignored.
    if ignore_template_numbers and type == "Number":
        return False
    # Only generate value for the first element of mutually exclusive inputs.
    if spec.xor and spec.xor[0] != name:
        return False
    # Directory types are not supported
    if "an existing directory name" in spec_info:
        return False
    # Don't know how to generate a list.
    if "a list" in spec_info or "a tuple" in spec_info:
        return False
    # Don't know how to generate a dictionary.
    if "a dictionary" in spec_info:
        return False
    # Best guess to detect string restrictions...
    if "' or '" in spec_info:
        return False
    if spec.default or spec.default_value():
        return False
    if not ignored_template_inputs:
        return True
    return not (name in ignored_template_inputs)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from .dti import create_bedpostx_pipeline, bedpostx_parallel

from .artifacts import (all_fmb_pipeline, all_peb_pipeline, all_fsl_pipeline,
                        hmc_pipeline, ecc_pipeline, sdc_fmb, sdc_peb,
                        remove_bias)

from .epi import (fieldmap_correction, topup_correction,
                  create_eddy_correct_pipeline, create_epidewarp_pipeline,
                  create_dmri_preprocessing)

from .tbss import (create_tbss_1_preproc, create_tbss_2_reg,
                   create_tbss_3_postreg, create_tbss_4_prestats,
                   create_tbss_all, create_tbss_non_FA)



# -*- coding: utf-8 -*-
# coding: utf-8
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import (print_function, division, unicode_literals,
                        absolute_import)

from ....interfaces.io import JSONFileGrabber
from ....interfaces import utility as niu
from ....interfaces import ants
from ....interfaces import fsl
from ....pipeline import engine as pe
from ...data import get_flirt_schedule

from .utils import (
    b0_indices,
    time_avg,
    apply_all_corrections,
    b0_average,
    hmc_split,
    dwi_flirt,
    eddy_rotate_bvecs,
    rotate_bvecs,
    insert_mat,
    extract_bval,
    recompose_dwi,
    recompose_xfm,
    siemens2rads,
    rads2radsec,
    demean_image,
    cleanup_edge_pipeline,
    add_empty_vol,
    vsm2warp,
    compute_readout,
)


def all_fmb_pipeline(name='hmc_sdc_ecc', fugue_params=dict(smooth3d=2.0)):
    """
    Builds a pipeline including three artifact corrections: head-motion
    correction (HMC), susceptibility-derived distortion correction (SDC),
    and Eddy currents-derived distortion correction (ECC).

    The displacement fields from each kind of distortions are combined. Thus,
    only one interpolation occurs between input data and result.

    .. warning:: this workflow rotates the gradients table (*b*-vectors)
      [Leemans09]_.


    Examples
    --------

    >>> from nipype.workflows.dmri.fsl.artifacts import all_fmb_pipeline
    >>> allcorr = all_fmb_pipeline()
    >>> allcorr.inputs.inputnode.in_file = 'epi.nii'
    >>> allcorr.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> allcorr.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> allcorr.inputs.inputnode.bmap_mag = 'magnitude.nii'
    >>> allcorr.inputs.inputnode.bmap_pha = 'phase.nii'
    >>> allcorr.inputs.inputnode.epi_param = 'epi_param.txt'
    >>> allcorr.run() # doctest: +SKIP

    """
    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file', 'in_bvec', 'in_bval', 'bmap_pha', 'bmap_mag',
            'epi_param'
        ]),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_mask', 'out_bvec']),
        name='outputnode')

    list_b0 = pe.Node(
        niu.Function(
            input_names=['in_bval'],
            output_names=['out_idx'],
            function=b0_indices),
        name='B0indices')

    avg_b0_0 = pe.Node(
        niu.Function(
            input_names=['in_file', 'index'],
            output_names=['out_file'],
            function=time_avg),
        name='b0_avg_pre')
    avg_b0_1 = pe.Node(
        niu.Function(
            input_names=['in_file', 'index'],
            output_names=['out_file'],
            function=time_avg),
        name='b0_avg_post')

    bet_dwi0 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_pre')
    bet_dwi1 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_post')

    hmc = hmc_pipeline()
    sdc = sdc_fmb(fugue_params=fugue_params)
    ecc = ecc_pipeline()
    unwarp = apply_all_corrections()

    wf = pe.Workflow(name=name)
    wf.connect(
        [(inputnode, hmc,
          [('in_file', 'inputnode.in_file'), ('in_bvec', 'inputnode.in_bvec'),
           ('in_bval', 'inputnode.in_bval')]), (inputnode, list_b0,
                                                [('in_bval', 'in_bval')]),
         (inputnode, avg_b0_0, [('in_file', 'in_file')]), (list_b0, avg_b0_0,
                                                           [('out_idx',
                                                             'index')]),
         (avg_b0_0, bet_dwi0, [('out_file', 'in_file')]), (bet_dwi0, hmc, [
             ('mask_file', 'inputnode.in_mask')
         ]), (hmc, sdc, [('outputnode.out_file', 'inputnode.in_file')]),
         (bet_dwi0, sdc,
          [('mask_file', 'inputnode.in_mask')]), (inputnode, sdc, [
              ('bmap_pha', 'inputnode.bmap_pha'),
              ('bmap_mag', 'inputnode.bmap_mag'), ('epi_param',
                                                   'inputnode.settings')
          ]), (list_b0, sdc, [('out_idx', 'inputnode.in_ref')]), (hmc, ecc, [
              ('outputnode.out_xfms', 'inputnode.in_xfms')
          ]), (inputnode, ecc,
               [('in_file', 'inputnode.in_file'),
                ('in_bval', 'inputnode.in_bval')]), (bet_dwi0, ecc, [
                    ('mask_file', 'inputnode.in_mask')
                ]), (ecc, avg_b0_1, [('outputnode.out_file',
                                      'in_file')]), (list_b0, avg_b0_1,
                                                     [('out_idx', 'index')]),
         (avg_b0_1, bet_dwi1, [('out_file', 'in_file')]), (inputnode, unwarp, [
             ('in_file', 'inputnode.in_dwi')
         ]), (hmc, unwarp,
              [('outputnode.out_xfms', 'inputnode.in_hmc')]), (ecc, unwarp, [
                  ('outputnode.out_xfms', 'inputnode.in_ecc')
              ]), (sdc, unwarp, [('outputnode.out_warp',
                                  'inputnode.in_sdc')]), (hmc, outputnode, [
                                      ('outputnode.out_bvec', 'out_bvec')
                                  ]), (unwarp, outputnode,
                                       [('outputnode.out_file',
                                         'out_file')]), (bet_dwi1, outputnode,
                                                         [('mask_file',
                                                           'out_mask')])])
    return wf


def all_peb_pipeline(name='hmc_sdc_ecc',
                     epi_params=dict(
                         echospacing=0.77e-3,
                         acc_factor=3,
                         enc_dir='y-',
                         epi_factor=1),
                     altepi_params=dict(
                         echospacing=0.77e-3,
                         acc_factor=3,
                         enc_dir='y',
                         epi_factor=1)):
    """
    Builds a pipeline including three artifact corrections: head-motion
    correction (HMC), susceptibility-derived distortion correction (SDC),
    and Eddy currents-derived distortion correction (ECC).

    .. warning:: this workflow rotates the gradients table (*b*-vectors)
      [Leemans09]_.


    Examples
    --------

    >>> from nipype.workflows.dmri.fsl.artifacts import all_peb_pipeline
    >>> allcorr = all_peb_pipeline()
    >>> allcorr.inputs.inputnode.in_file = 'epi.nii'
    >>> allcorr.inputs.inputnode.alt_file = 'epi_rev.nii'
    >>> allcorr.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> allcorr.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> allcorr.run() # doctest: +SKIP

    """
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_file', 'in_bvec', 'in_bval', 'alt_file']),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_mask', 'out_bvec']),
        name='outputnode')

    avg_b0_0 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg_pre')
    avg_b0_1 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg_post')
    bet_dwi0 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_pre')
    bet_dwi1 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_post')

    hmc = hmc_pipeline()
    sdc = sdc_peb(epi_params=epi_params, altepi_params=altepi_params)
    ecc = ecc_pipeline()

    unwarp = apply_all_corrections()

    wf = pe.Workflow(name=name)
    wf.connect(
        [(inputnode, hmc,
          [('in_file', 'inputnode.in_file'), ('in_bvec', 'inputnode.in_bvec'),
           ('in_bval', 'inputnode.in_bval')]), (inputnode, avg_b0_0,
                                                [('in_file', 'in_dwi'),
                                                 ('in_bval', 'in_bval')]),
         (avg_b0_0, bet_dwi0, [('out_file', 'in_file')]), (bet_dwi0, hmc, [
             ('mask_file', 'inputnode.in_mask')
         ]), (hmc, sdc, [('outputnode.out_file', 'inputnode.in_file')]),
         (bet_dwi0, sdc,
          [('mask_file', 'inputnode.in_mask')]), (inputnode, sdc, [
              ('in_bval', 'inputnode.in_bval'), ('alt_file',
                                                 'inputnode.alt_file')
          ]), (inputnode, ecc, [('in_file', 'inputnode.in_file'),
                                ('in_bval', 'inputnode.in_bval')]),
         (bet_dwi0, ecc, [('mask_file', 'inputnode.in_mask')]), (hmc, ecc, [
             ('outputnode.out_xfms', 'inputnode.in_xfms')
         ]), (ecc, avg_b0_1, [('outputnode.out_file',
                               'in_dwi')]), (inputnode, avg_b0_1,
                                             [('in_bval', 'in_bval')]),
         (avg_b0_1, bet_dwi1, [('out_file', 'in_file')]), (inputnode, unwarp, [
             ('in_file', 'inputnode.in_dwi')
         ]), (hmc, unwarp,
              [('outputnode.out_xfms', 'inputnode.in_hmc')]), (ecc, unwarp, [
                  ('outputnode.out_xfms', 'inputnode.in_ecc')
              ]), (sdc, unwarp, [('outputnode.out_warp',
                                  'inputnode.in_sdc')]), (hmc, outputnode, [
                                      ('outputnode.out_bvec', 'out_bvec')
                                  ]), (unwarp, outputnode,
                                       [('outputnode.out_file',
                                         'out_file')]), (bet_dwi1, outputnode,
                                                         [('mask_file',
                                                           'out_mask')])])
    return wf


def all_fsl_pipeline(name='fsl_all_correct',
                     epi_params=dict(
                         echospacing=0.77e-3, acc_factor=3, enc_dir='y-'),
                     altepi_params=dict(
                         echospacing=0.77e-3, acc_factor=3, enc_dir='y')):
    """
    Workflow that integrates FSL ``topup`` and ``eddy``.


    .. warning:: this workflow rotates the gradients table (*b*-vectors)
      [Leemans09]_.


    .. warning:: this workflow does not perform jacobian modulation of each
      *DWI* [Jones10]_.


    Examples
    --------

    >>> from nipype.workflows.dmri.fsl.artifacts import all_fsl_pipeline
    >>> allcorr = all_fsl_pipeline()
    >>> allcorr.inputs.inputnode.in_file = 'epi.nii'
    >>> allcorr.inputs.inputnode.alt_file = 'epi_rev.nii'
    >>> allcorr.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> allcorr.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> allcorr.run() # doctest: +SKIP

    """

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_file', 'in_bvec', 'in_bval', 'alt_file']),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_mask', 'out_bvec']),
        name='outputnode')

    def gen_index(in_file):
        import numpy as np
        import nibabel as nb
        import os
        from nipype.utils import NUMPY_MMAP
        out_file = os.path.abspath('index.txt')
        vols = nb.load(in_file, mmap=NUMPY_MMAP).get_data().shape[-1]
        np.savetxt(out_file, np.ones((vols, )).T)
        return out_file

    gen_idx = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=gen_index),
        name='gen_index')
    avg_b0_0 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg_pre')
    bet_dwi0 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_pre')

    sdc = sdc_peb(epi_params=epi_params, altepi_params=altepi_params)
    ecc = pe.Node(fsl.Eddy(method='jac'), name='fsl_eddy')
    rot_bvec = pe.Node(
        niu.Function(
            input_names=['in_bvec', 'eddy_params'],
            output_names=['out_file'],
            function=eddy_rotate_bvecs),
        name='Rotate_Bvec')
    avg_b0_1 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg_post')
    bet_dwi1 = pe.Node(
        fsl.BET(frac=0.3, mask=True, robust=True), name='bet_dwi_post')

    wf = pe.Workflow(name=name)
    wf.connect(
        [(inputnode, avg_b0_0, [('in_file', 'in_dwi'), ('in_bval',
                                                        'in_bval')]),
         (avg_b0_0, bet_dwi0, [('out_file', 'in_file')]), (bet_dwi0, sdc, [
             ('mask_file', 'inputnode.in_mask')
         ]), (inputnode, sdc, [('in_file', 'inputnode.in_file'),
                               ('alt_file', 'inputnode.alt_file'),
                               ('in_bval', 'inputnode.in_bval')]),
         (sdc, ecc, [('topup.out_enc_file', 'in_acqp'),
                     ('topup.out_fieldcoef', 'in_topup_fieldcoef'),
                     ('topup.out_movpar',
                      'in_topup_movpar')]), (bet_dwi0, ecc, [('mask_file',
                                                              'in_mask')]),
         (inputnode, gen_idx, [('in_file', 'in_file')]), (inputnode, ecc, [
             ('in_file', 'in_file'), ('in_bval', 'in_bval'), ('in_bvec',
                                                              'in_bvec')
         ]), (gen_idx, ecc,
              [('out_file', 'in_index')]), (inputnode, rot_bvec, [
                  ('in_bvec', 'in_bvec')
              ]), (ecc, rot_bvec,
                   [('out_parameter', 'eddy_params')]), (ecc, avg_b0_1, [
                       ('out_corrected', 'in_dwi')
                   ]), (inputnode, avg_b0_1, [('in_bval', 'in_bval')]),
         (avg_b0_1, bet_dwi1, [('out_file', 'in_file')]), (ecc, outputnode, [
             ('out_corrected', 'out_file')
         ]), (rot_bvec, outputnode,
              [('out_file', 'out_bvec')]), (bet_dwi1, outputnode,
                                            [('mask_file', 'out_mask')])])
    return wf


def hmc_pipeline(name='motion_correct'):
    """
    HMC stands for head-motion correction.

    Creates a pipeline that corrects for head motion artifacts in dMRI
    sequences.
    It takes a series of diffusion weighted images and rigidly co-registers
    them to one reference image. Finally, the `b`-matrix is rotated accordingly
    [Leemans09]_ making use of the rotation matrix obtained by FLIRT.

    Search angles have been limited to 4 degrees, based on results in
    [Yendiki13]_.

    A list of rigid transformation matrices is provided, so that transforms
    can be chained.
    This is useful to correct for artifacts with only one interpolation process
    (as previously discussed `here
    <https://github.com/nipy/nipype/pull/530#issuecomment-14505042>`_),
    and also to compute nuisance regressors as proposed by [Yendiki13]_.

    .. warning:: This workflow rotates the `b`-vectors, so please be advised
      that not all the dicom converters ensure the consistency between the
      resulting nifti orientation and the gradients table (e.g. dcm2nii
      checks it).

    .. admonition:: References

      .. [Leemans09] Leemans A, and Jones DK, `The B-matrix must be rotated
        when correcting for subject motion in DTI data
        <http://dx.doi.org/10.1002/mrm.21890>`_,
        Magn Reson Med. 61(6):1336-49. 2009. doi: 10.1002/mrm.21890.

      .. [Yendiki13] Yendiki A et al., `Spurious group differences due to head
        motion in a diffusion MRI study
        <http://dx.doi.org/10.1016/j.neuroimage.2013.11.027>`_.
        Neuroimage. 21(88C):79-90. 2013. doi: 10.1016/j.neuroimage.2013.11.027

    Example
    -------

    >>> from nipype.workflows.dmri.fsl.artifacts import hmc_pipeline
    >>> hmc = hmc_pipeline()
    >>> hmc.inputs.inputnode.in_file = 'diffusion.nii'
    >>> hmc.inputs.inputnode.in_bvec = 'diffusion.bvec'
    >>> hmc.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> hmc.inputs.inputnode.in_mask = 'mask.nii'
    >>> hmc.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file - input dwi file
        inputnode.in_mask - weights mask of reference image (a file with data \
range in [0.0, 1.0], indicating the weight of each voxel when computing the \
metric.
        inputnode.in_bval - b-values file
        inputnode.in_bvec - gradients file (b-vectors)
        inputnode.ref_num (optional, default=0) index of the b0 volume that \
should be taken as reference

    Outputs::

        outputnode.out_file - corrected dwi file
        outputnode.out_bvec - rotated gradient vectors table
        outputnode.out_xfms - list of transformation matrices

    """
    params = dict(
        dof=6,
        bgvalue=0,
        save_log=True,
        no_search=True,
        # cost='mutualinfo', cost_func='mutualinfo', bins=64,
        schedule=get_flirt_schedule('hmc'))

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_file', 'ref_num', 'in_bvec', 'in_bval', 'in_mask']),
        name='inputnode')
    split = pe.Node(
        niu.Function(
            output_names=['out_ref', 'out_mov', 'out_bval', 'volid'],
            input_names=['in_file', 'in_bval', 'ref_num'],
            function=hmc_split),
        name='SplitDWI')
    flirt = dwi_flirt(flirt_param=params)
    insmat = pe.Node(
        niu.Function(
            input_names=['inlist', 'volid'],
            output_names=['out'],
            function=insert_mat),
        name='InsertRefmat')
    rot_bvec = pe.Node(
        niu.Function(
            function=rotate_bvecs,
            input_names=['in_bvec', 'in_matrix'],
            output_names=['out_file']),
        name='Rotate_Bvec')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_bvec', 'out_xfms']),
        name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, split,
                 [('in_file', 'in_file'), ('in_bval', 'in_bval'),
                  ('ref_num', 'ref_num')]), (inputnode, flirt, [
                      ('in_mask', 'inputnode.ref_mask')
                  ]), (split, flirt, [('out_ref', 'inputnode.reference'),
                                      ('out_mov', 'inputnode.in_file'),
                                      ('out_bval', 'inputnode.in_bval')]),
                (flirt, insmat, [('outputnode.out_xfms', 'inlist')]),
                (split, insmat, [('volid', 'volid')]), (inputnode, rot_bvec, [
                    ('in_bvec', 'in_bvec')
                ]), (insmat, rot_bvec,
                     [('out', 'in_matrix')]), (rot_bvec, outputnode,
                                               [('out_file', 'out_bvec')]),
                (flirt, outputnode, [('outputnode.out_file',
                                      'out_file')]), (insmat, outputnode,
                                                      [('out', 'out_xfms')])])
    return wf


def ecc_pipeline(name='eddy_correct'):
    """
    ECC stands for Eddy currents correction.

    Creates a pipeline that corrects for artifacts induced by Eddy currents in
    dMRI sequences.
    It takes a series of diffusion weighted images and linearly co-registers
    them to one reference image (the average of all b0s in the dataset).

    DWIs are also modulated by the determinant of the Jacobian as indicated by
    [Jones10]_ and [Rohde04]_.

    A list of rigid transformation matrices can be provided, sourcing from a
    :func:`.hmc_pipeline` workflow, to initialize registrations in a *motion
    free* framework.

    A list of affine transformation matrices is available as output, so that
    transforms can be chained (discussion
    `here <https://github.com/nipy/nipype/pull/530#issuecomment-14505042>`_).

    .. admonition:: References

      .. [Jones10] Jones DK, `The signal intensity must be modulated by the
        determinant of the Jacobian when correcting for eddy currents in
        diffusion MRI
        <http://cds.ismrm.org/protected/10MProceedings/files/1644_129.pdf>`_,
        Proc. ISMRM 18th Annual Meeting, (2010).

      .. [Rohde04] Rohde et al., `Comprehensive Approach for Correction of
        Motion and Distortion in Diffusion-Weighted MRI
        <http://stbb.nichd.nih.gov/pdf/com_app_cor_mri04.pdf>`_, MRM
        51:103-114 (2004).

    Example
    -------

    >>> from nipype.workflows.dmri.fsl.artifacts import ecc_pipeline
    >>> ecc = ecc_pipeline()
    >>> ecc.inputs.inputnode.in_file = 'diffusion.nii'
    >>> ecc.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> ecc.inputs.inputnode.in_mask = 'mask.nii'
    >>> ecc.run() # doctest: +SKIP

    Inputs::

        inputnode.in_file - input dwi file
        inputnode.in_mask - weights mask of reference image (a file with data \
range sin [0.0, 1.0], indicating the weight of each voxel when computing the \
metric.
        inputnode.in_bval - b-values table
        inputnode.in_xfms - list of matrices to initialize registration (from \
head-motion correction)

    Outputs::

        outputnode.out_file - corrected dwi file
        outputnode.out_xfms - list of transformation matrices
    """

    params = dict(
        dof=12,
        no_search=True,
        interp='spline',
        bgvalue=0,
        schedule=get_flirt_schedule('ecc'))
    # cost='normmi', cost_func='normmi', bins=64,

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_file', 'in_bval', 'in_mask', 'in_xfms']),
        name='inputnode')
    avg_b0 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg')
    pick_dws = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval', 'b'],
            output_names=['out_file'],
            function=extract_bval),
        name='ExtractDWI')
    pick_dws.inputs.b = 'diff'

    flirt = dwi_flirt(flirt_param=params, excl_nodiff=True)

    mult = pe.MapNode(
        fsl.BinaryMaths(operation='mul'),
        name='ModulateDWIs',
        iterfield=['in_file', 'operand_value'])
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=['in_file'],
        name='RemoveNegative')

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    get_mat = pe.Node(
        niu.Function(
            input_names=['in_bval', 'in_xfms'],
            output_names=['out_files'],
            function=recompose_xfm),
        name='GatherMatrices')
    merge = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval', 'in_corrected'],
            output_names=['out_file'],
            function=recompose_dwi),
        name='MergeDWIs')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_xfms']),
        name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, avg_b0, [('in_file', 'in_dwi'), ('in_bval', 'in_bval')]),
        (inputnode, pick_dws, [('in_file', 'in_dwi'), ('in_bval', 'in_bval')]),
        (inputnode, merge,
         [('in_file', 'in_dwi'), ('in_bval', 'in_bval')]), (inputnode, flirt, [
             ('in_mask', 'inputnode.ref_mask'),
             ('in_xfms', 'inputnode.in_xfms'), ('in_bval', 'inputnode.in_bval')
         ]), (inputnode, get_mat, [('in_bval', 'in_bval')]), (avg_b0, flirt, [
             ('out_file', 'inputnode.reference')
         ]), (pick_dws, flirt, [('out_file', 'inputnode.in_file')]),
        (flirt, get_mat, [('outputnode.out_xfms', 'in_xfms')]), (flirt, mult, [
            (('outputnode.out_xfms', _xfm_jacobian), 'operand_value')
        ]), (flirt, split,
             [('outputnode.out_file', 'in_file')]), (split, mult, [
                 ('out_files', 'in_file')
             ]), (mult, thres, [('out_file', 'in_file')]), (thres, merge, [
                 ('out_file', 'in_corrected')
             ]), (get_mat, outputnode,
                  [('out_files', 'out_xfms')]), (merge, outputnode,
                                                 [('out_file', 'out_file')])
    ])
    return wf


def sdc_fmb(name='fmb_correction',
            interp='Linear',
            fugue_params=dict(smooth3d=2.0)):
    """
    SDC stands for susceptibility distortion correction. FMB stands for
    fieldmap-based.

    The fieldmap based (FMB) method implements SDC by using a mapping of the
    B0 field as proposed by [Jezzard95]_. This workflow uses the implementation
    of FSL (`FUGUE <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE>`_). Phase
    unwrapping is performed using `PRELUDE
    <http://fsl.fmrib.ox.ac.uk/fsl/fsl-4.1.9/fugue/prelude.html>`_
    [Jenkinson03]_. Preparation of the fieldmap is performed reproducing the
    script in FSL `fsl_prepare_fieldmap
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE/Guide#SIEMENS_data>`_.



    Example
    -------

    >>> from nipype.workflows.dmri.fsl.artifacts import sdc_fmb
    >>> fmb = sdc_fmb()
    >>> fmb.inputs.inputnode.in_file = 'diffusion.nii'
    >>> fmb.inputs.inputnode.in_ref = list(range(0, 30, 6))
    >>> fmb.inputs.inputnode.in_mask = 'mask.nii'
    >>> fmb.inputs.inputnode.bmap_mag = 'magnitude.nii'
    >>> fmb.inputs.inputnode.bmap_pha = 'phase.nii'
    >>> fmb.inputs.inputnode.settings = 'epi_param.txt'
    >>> fmb.run() # doctest: +SKIP

    .. warning:: Only SIEMENS format fieldmaps are supported.

    .. admonition:: References

      .. [Jezzard95] Jezzard P, and Balaban RS, `Correction for geometric
        distortion in echo planar images from B0 field variations
        <http://dx.doi.org/10.1002/mrm.1910340111>`_,
        MRM 34(1):65-73. (1995). doi: 10.1002/mrm.1910340111.

      .. [Jenkinson03] Jenkinson M., `Fast, automated, N-dimensional
        phase-unwrapping algorithm <http://dx.doi.org/10.1002/mrm.10354>`_,
        MRM 49(1):193-197, 2003, doi: 10.1002/mrm.10354.

    """

    epi_defaults = {
        'delta_te': 2.46e-3,
        'echospacing': 0.77e-3,
        'acc_factor': 2,
        'enc_dir': u'AP'
    }

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[
            'in_file', 'in_ref', 'in_mask', 'bmap_pha', 'bmap_mag', 'settings'
        ]),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_vsm', 'out_warp']),
        name='outputnode')

    r_params = pe.Node(
        JSONFileGrabber(defaults=epi_defaults), name='SettingsGrabber')
    eff_echo = pe.Node(
        niu.Function(
            function=_eff_t_echo,
            input_names=['echospacing', 'acc_factor'],
            output_names=['eff_echo']),
        name='EffEcho')

    firstmag = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='GetFirst')
    n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='Bias')
    bet = pe.Node(fsl.BET(frac=0.4, mask=True), name='BrainExtraction')
    dilate = pe.Node(
        fsl.maths.MathsCommand(nan2zeros=True, args='-kernel sphere 5 -dilM'),
        name='MskDilate')
    pha2rads = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=siemens2rads),
        name='PreparePhase')
    prelude = pe.Node(fsl.PRELUDE(process3d=True), name='PhaseUnwrap')
    rad2rsec = pe.Node(
        niu.Function(
            input_names=['in_file', 'delta_te'],
            output_names=['out_file'],
            function=rads2radsec),
        name='ToRadSec')

    baseline = pe.Node(
        niu.Function(
            input_names=['in_file', 'index'],
            output_names=['out_file'],
            function=time_avg),
        name='Baseline')

    fmm2b0 = pe.Node(
        ants.Registration(output_warped_image=True), name="FMm_to_B0")
    fmm2b0.inputs.transforms = ['Rigid'] * 2
    fmm2b0.inputs.transform_parameters = [(1.0, )] * 2
    fmm2b0.inputs.number_of_iterations = [[50], [20]]
    fmm2b0.inputs.dimension = 3
    fmm2b0.inputs.metric = ['Mattes', 'Mattes']
    fmm2b0.inputs.metric_weight = [1.0] * 2
    fmm2b0.inputs.radius_or_number_of_bins = [64, 64]
    fmm2b0.inputs.sampling_strategy = ['Regular', 'Random']
    fmm2b0.inputs.sampling_percentage = [None, 0.2]
    fmm2b0.inputs.convergence_threshold = [1.e-5, 1.e-8]
    fmm2b0.inputs.convergence_window_size = [20, 10]
    fmm2b0.inputs.smoothing_sigmas = [[6.0], [2.0]]
    fmm2b0.inputs.sigma_units = ['vox'] * 2
    fmm2b0.inputs.shrink_factors = [[6], [1]]  # ,[1] ]
    fmm2b0.inputs.use_estimate_learning_rate_once = [True] * 2
    fmm2b0.inputs.use_histogram_matching = [True] * 2
    fmm2b0.inputs.initial_moving_transform_com = 0
    fmm2b0.inputs.collapse_output_transforms = True
    fmm2b0.inputs.winsorize_upper_quantile = 0.995

    applyxfm = pe.Node(
        ants.ApplyTransforms(dimension=3, interpolation=interp),
        name='FMp_to_B0')

    pre_fugue = pe.Node(fsl.FUGUE(save_fmap=True), name='PreliminaryFugue')
    demean = pe.Node(
        niu.Function(
            input_names=['in_file', 'in_mask'],
            output_names=['out_file'],
            function=demean_image),
        name='DemeanFmap')

    cleanup = cleanup_edge_pipeline()

    addvol = pe.Node(
        niu.Function(
            input_names=['in_file'],
            output_names=['out_file'],
            function=add_empty_vol),
        name='AddEmptyVol')

    vsm = pe.Node(
        fsl.FUGUE(save_shift=True, **fugue_params), name="ComputeVSM")

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')
    unwarp = pe.MapNode(
        fsl.FUGUE(icorr=True, forward_warping=False),
        iterfield=['in_file'],
        name='UnwarpDWIs')
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=['in_file'],
        name='RemoveNegative')
    vsm2dfm = vsm2warp()
    vsm2dfm.inputs.inputnode.scaling = 1.0

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, r_params,
         [('settings', 'in_file')]), (r_params, eff_echo, [
             ('echospacing', 'echospacing'), ('acc_factor', 'acc_factor')
         ]), (inputnode, pha2rads,
              [('bmap_pha', 'in_file')]), (inputnode, firstmag,
                                           [('bmap_mag', 'in_file')]),
        (inputnode, baseline,
         [('in_file', 'in_file'), ('in_ref', 'index')]), (firstmag, n4, [
             ('roi_file', 'input_image')
         ]), (n4, bet, [('output_image', 'in_file')]), (bet, dilate, [
             ('mask_file', 'in_file')
         ]), (pha2rads, prelude, [('out_file', 'phase_file')]), (n4, prelude, [
             ('output_image', 'magnitude_file')
         ]), (dilate, prelude, [('out_file', 'mask_file')]),
        (r_params, rad2rsec, [('delta_te', 'delta_te')]), (prelude, rad2rsec, [
            ('unwrapped_phase_file', 'in_file')
        ]), (baseline, fmm2b0, [('out_file', 'fixed_image')]), (n4, fmm2b0, [
            ('output_image', 'moving_image')
        ]), (inputnode, fmm2b0,
             [('in_mask', 'fixed_image_mask')]), (dilate, fmm2b0, [
                 ('out_file', 'moving_image_mask')
             ]), (baseline, applyxfm, [('out_file', 'reference_image')]),
        (rad2rsec, applyxfm,
         [('out_file', 'input_image')]), (fmm2b0, applyxfm, [
             ('forward_transforms', 'transforms'), ('forward_invert_flags',
                                                    'invert_transform_flags')
         ]), (applyxfm, pre_fugue,
              [('output_image', 'fmap_in_file')]), (inputnode, pre_fugue, [
                  ('in_mask', 'mask_file')
              ]), (pre_fugue, demean,
                   [('fmap_out_file', 'in_file')]), (inputnode, demean, [
                       ('in_mask', 'in_mask')
                   ]), (demean, cleanup, [('out_file', 'inputnode.in_file')]),
        (inputnode, cleanup,
         [('in_mask', 'inputnode.in_mask')]), (cleanup, addvol, [
             ('outputnode.out_file', 'in_file')
         ]), (inputnode, vsm, [('in_mask', 'mask_file')]), (addvol, vsm, [
             ('out_file', 'fmap_in_file')
         ]), (r_params, vsm, [('delta_te', 'asym_se_time')]), (eff_echo, vsm, [
             ('eff_echo', 'dwell_time')
         ]), (inputnode, split, [('in_file', 'in_file')]), (split, unwarp, [
             ('out_files', 'in_file')
         ]), (vsm, unwarp,
              [('shift_out_file', 'shift_in_file')]), (r_params, unwarp, [
                  (('enc_dir', _fix_enc_dir), 'unwarp_direction')
              ]), (unwarp, thres, [('unwarped_file', 'in_file')]),
        (thres, merge, [('out_file', 'in_files')]), (r_params, vsm2dfm, [
            (('enc_dir', _fix_enc_dir), 'inputnode.enc_dir')
        ]), (merge, vsm2dfm,
             [('merged_file', 'inputnode.in_ref')]), (vsm, vsm2dfm, [
                 ('shift_out_file', 'inputnode.in_vsm')
             ]), (merge, outputnode,
                  [('merged_file', 'out_file')]), (vsm, outputnode, [
                      ('shift_out_file', 'out_vsm')
                  ]), (vsm2dfm, outputnode, [('outputnode.out_warp',
                                              'out_warp')])
    ])
    return wf


def sdc_peb(name='peb_correction',
            epi_params=dict(
                echospacing=0.77e-3, acc_factor=3, enc_dir='y-', epi_factor=1),
            altepi_params=dict(
                echospacing=0.77e-3, acc_factor=3, enc_dir='y', epi_factor=1)):
    """
    SDC stands for susceptibility distortion correction. PEB stands for
    phase-encoding-based.

    The phase-encoding-based (PEB) method implements SDC by acquiring
    diffusion images with two different enconding directions [Andersson2003]_.
    The most typical case is acquiring with opposed phase-gradient blips
    (e.g. *A>>>P* and *P>>>A*, or equivalently, *-y* and *y*)
    as in [Chiou2000]_, but it is also possible to use orthogonal
    configurations [Cordes2000]_ (e.g. *A>>>P* and *L>>>R*,
    or equivalently *-y* and *x*).
    This workflow uses the implementation of FSL
    (`TOPUP <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/TOPUP>`_).

    Example
    -------

    >>> from nipype.workflows.dmri.fsl.artifacts import sdc_peb
    >>> peb = sdc_peb()
    >>> peb.inputs.inputnode.in_file = 'epi.nii'
    >>> peb.inputs.inputnode.alt_file = 'epi_rev.nii'
    >>> peb.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> peb.inputs.inputnode.in_mask = 'mask.nii'
    >>> peb.run() # doctest: +SKIP

    .. admonition:: References

      .. [Andersson2003] Andersson JL et al., `How to correct susceptibility
        distortions in spin-echo echo-planar images: application to diffusion
        tensor imaging <http://dx.doi.org/10.1016/S1053-8119(03)00336-7>`_.
        Neuroimage. 2003 Oct;20(2):870-88. doi: 10.1016/S1053-8119(03)00336-7

      .. [Cordes2000] Cordes D et al., Geometric distortion correction in EPI
        using two images with orthogonal phase-encoding directions, in Proc.
        ISMRM (8), p.1712, Denver, US, 2000.

      .. [Chiou2000] Chiou JY, and Nalcioglu O, A simple method to correct
        off-resonance related distortion in echo planar imaging, in Proc.
        ISMRM (8), p.1712, Denver, US, 2000.

    """

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['in_file', 'in_bval', 'in_mask', 'alt_file', 'ref_num']),
        name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file', 'out_vsm', 'out_warp']),
        name='outputnode')

    b0_ref = pe.Node(fsl.ExtractROI(t_size=1), name='b0_ref')
    b0_alt = pe.Node(fsl.ExtractROI(t_size=1), name='b0_alt')
    b0_comb = pe.Node(niu.Merge(2), name='b0_list')
    b0_merge = pe.Node(fsl.Merge(dimension='t'), name='b0_merged')

    topup = pe.Node(fsl.TOPUP(), name='topup')
    topup.inputs.encoding_direction = [
        epi_params['enc_dir'], altepi_params['enc_dir']
    ]

    readout = compute_readout(epi_params)
    topup.inputs.readout_times = [readout, compute_readout(altepi_params)]

    unwarp = pe.Node(fsl.ApplyTOPUP(in_index=[1], method='jac'), name='unwarp')

    # scaling = pe.Node(niu.Function(input_names=['in_file', 'enc_dir'],
    #                   output_names=['factor'], function=_get_zoom),
    #                   name='GetZoom')
    # scaling.inputs.enc_dir = epi_params['enc_dir']
    vsm2dfm = vsm2warp()
    vsm2dfm.inputs.inputnode.enc_dir = epi_params['enc_dir']
    vsm2dfm.inputs.inputnode.scaling = readout

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, b0_ref, [('in_file', 'in_file'), (('ref_num', _checkrnum),
                                                      't_min')]),
        (inputnode, b0_alt, [('alt_file', 'in_file'), (('ref_num', _checkrnum),
                                                       't_min')]),
        (b0_ref, b0_comb, [('roi_file', 'in1')]),
        (b0_alt, b0_comb, [('roi_file', 'in2')]),
        (b0_comb, b0_merge, [('out', 'in_files')]),
        (b0_merge, topup, [('merged_file', 'in_file')]),
        (topup, unwarp, [('out_fieldcoef', 'in_topup_fieldcoef'),
                         ('out_movpar', 'in_topup_movpar'),
                         ('out_enc_file', 'encoding_file')]),
        (inputnode, unwarp, [('in_file', 'in_files')]),
        (unwarp, outputnode, [('out_corrected', 'out_file')]),
        # (b0_ref,      scaling,    [('roi_file', 'in_file')]),
        # (scaling,     vsm2dfm,    [('factor', 'inputnode.scaling')]),
        (b0_ref, vsm2dfm, [('roi_file', 'inputnode.in_ref')]),
        (topup, vsm2dfm, [('out_field', 'inputnode.in_vsm')]),
        (topup, outputnode, [('out_field', 'out_vsm')]),
        (vsm2dfm, outputnode, [('outputnode.out_warp', 'out_warp')])
    ])
    return wf


def remove_bias(name='bias_correct'):
    """
    This workflow estimates a single multiplicative bias field from the
    averaged *b0* image, as suggested in [Jeurissen2014]_.

    .. admonition:: References

      .. [Jeurissen2014] Jeurissen B. et al., `Multi-tissue constrained
        spherical deconvolution for improved analysis of multi-shell diffusion
        MRI data <http://dx.doi.org/10.1016/j.neuroimage.2014.07.061>`_.
        NeuroImage (2014). doi: 10.1016/j.neuroimage.2014.07.061


    Example
    -------

    >>> from nipype.workflows.dmri.fsl.artifacts import remove_bias
    >>> bias = remove_bias()
    >>> bias.inputs.inputnode.in_file = 'epi.nii'
    >>> bias.inputs.inputnode.in_bval = 'diffusion.bval'
    >>> bias.inputs.inputnode.in_mask = 'mask.nii'
    >>> bias.run() # doctest: +SKIP

    """
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_file', 'in_bval', 'in_mask']),
        name='inputnode')

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['out_file']), name='outputnode')

    avg_b0 = pe.Node(
        niu.Function(
            input_names=['in_dwi', 'in_bval'],
            output_names=['out_file'],
            function=b0_average),
        name='b0_avg')
    n4 = pe.Node(
        ants.N4BiasFieldCorrection(
            dimension=3, save_bias=True, bspline_fitting_distance=600),
        name='Bias_b0')
    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')
    mult = pe.MapNode(
        fsl.MultiImageMaths(op_string='-div %s'),
        iterfield=['in_file'],
        name='RemoveBiasOfDWIs')
    thres = pe.MapNode(
        fsl.Threshold(thresh=0.0),
        iterfield=['in_file'],
        name='RemoveNegative')
    merge = pe.Node(fsl.utils.Merge(dimension='t'), name='MergeDWIs')

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, avg_b0, [
        ('in_file', 'in_dwi'), ('in_bval', 'in_bval')
    ]), (avg_b0, n4, [('out_file', 'input_image')]), (inputnode, n4, [
        ('in_mask', 'mask_image')
    ]), (inputnode, split, [('in_file', 'in_file')]), (n4, mult, [
        ('bias_image', 'operand_files')
    ]), (split, mult, [('out_files', 'in_file')]), (mult, thres,
                                                    [('out_file', 'in_file')]),
                (thres, merge, [('out_file', 'in_files')]),
                (merge, outputnode, [('merged_file', 'out_file')])])
    return wf


def _eff_t_echo(echospacing, acc_factor):
    eff_echo = echospacing / (1.0 * acc_factor)
    return eff_echo


def _fix_enc_dir(enc_dir):
    enc_dir = enc_dir.lower()
    if enc_dir == 'lr':
        return 'x-'
    if enc_dir == 'rl':
        return 'x'
    if enc_dir == 'ap':
        return 'y-'
    if enc_dir == 'pa':
        return 'y'
    return enc_dir


def _checkrnum(ref_num):
    from nipype.interfaces.base import isdefined
    if (ref_num is None) or not isdefined(ref_num):
        return 0
    return ref_num


def _nonb0(in_bval):
    import numpy as np
    bvals = np.loadtxt(in_bval)
    return np.where(bvals != 0)[0].tolist()


def _xfm_jacobian(in_xfm):
    import numpy as np
    from math import fabs
    return [fabs(np.linalg.det(np.loadtxt(xfm))) for xfm in in_xfm]


def _get_zoom(in_file, enc_dir):
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP

    zooms = nb.load(in_file, mmap=NUMPY_MMAP).header.get_zooms()

    if 'y' in enc_dir:
        return zooms[1]
    elif 'x' in enc_dir:
        return zooms[0]
    elif 'z' in enc_dir:
        return zooms[2]
    else:
        raise ValueError('Wrong encoding direction string')



# -*- coding: utf-8 -*-
from ....interfaces import utility as util  # utility
from ....pipeline import engine as pe  # pypeline engine
from ....interfaces import camino as camino
from ....interfaces import fsl as fsl
from ....interfaces import camino2trackvis as cam2trk
from ....algorithms import misc as misc
from ...misc.utils import get_affine, get_data_dims, get_vox_dims


def create_camino_dti_pipeline(name="dtiproc"):
    """Creates a pipeline that does the same diffusion processing as in the
    :doc:`../../users/examples/dmri_camino_dti` example script. Given a diffusion-weighted image,
    b-values, and b-vectors, the workflow will return the tractography
    computed from diffusion tensors and from PICo probabilistic tractography.

    Example
    -------

    >>> import os
    >>> nipype_camino_dti = create_camino_dti_pipeline("nipype_camino_dti")
    >>> nipype_camino_dti.inputs.inputnode.dwi = os.path.abspath('dwi.nii')
    >>> nipype_camino_dti.inputs.inputnode.bvecs = os.path.abspath('bvecs')
    >>> nipype_camino_dti.inputs.inputnode.bvals = os.path.abspath('bvals')
    >>> nipype_camino_dti.run()                  # doctest: +SKIP

    Inputs::

        inputnode.dwi
        inputnode.bvecs
        inputnode.bvals

    Outputs::

        outputnode.fa
        outputnode.trace
        outputnode.tracts_pico
        outputnode.tracts_dt
        outputnode.tensors

    """

    inputnode1 = pe.Node(
        interface=util.IdentityInterface(fields=["dwi", "bvecs", "bvals"]),
        name="inputnode1")
    """
    Setup for Diffusion Tensor Computation
    --------------------------------------
    In this section we create the nodes necessary for diffusion analysis.
    First, the diffusion image is converted to voxel order.
    """

    image2voxel = pe.Node(interface=camino.Image2Voxel(), name="image2voxel")
    fsl2scheme = pe.Node(interface=camino.FSL2Scheme(), name="fsl2scheme")
    fsl2scheme.inputs.usegradmod = True
    """
    Second, diffusion tensors are fit to the voxel-order data.
    """

    dtifit = pe.Node(interface=camino.DTIFit(), name='dtifit')
    """
    Next, a lookup table is generated from the schemefile and the
    signal-to-noise ratio (SNR) of the unweighted (q=0) data.
    """

    dtlutgen = pe.Node(interface=camino.DTLUTGen(), name="dtlutgen")
    dtlutgen.inputs.snr = 16.0
    dtlutgen.inputs.inversion = 1
    """
    In this tutorial we implement probabilistic tractography using the PICo algorithm.
    PICo tractography requires an estimate of the fibre direction and a model of its
    uncertainty in each voxel; this is produced using the following node.
    """

    picopdfs = pe.Node(interface=camino.PicoPDFs(), name="picopdfs")
    picopdfs.inputs.inputmodel = 'dt'
    """
    An FSL BET node creates a brain mask is generated from the diffusion image for seeding the PICo     tractography.
    """

    bet = pe.Node(interface=fsl.BET(), name="bet")
    bet.inputs.mask = True
    """
    Finally, tractography is performed.
    First DT streamline tractography.
    """

    trackdt = pe.Node(interface=camino.TrackDT(), name="trackdt")
    """
    Now camino's Probablistic Index of connectivity algorithm.
    In this tutorial, we will use only 1 iteration for time-saving purposes.
    """

    trackpico = pe.Node(interface=camino.TrackPICo(), name="trackpico")
    trackpico.inputs.iterations = 1
    """
    Currently, the best program for visualizing tracts is TrackVis. For this reason, a node is included         to convert the raw tract data to .trk format. Solely for testing purposes, another node is added to         perform the reverse.
    """

    cam2trk_dt = pe.Node(
        interface=cam2trk.Camino2Trackvis(), name="cam2trk_dt")
    cam2trk_dt.inputs.min_length = 30
    cam2trk_dt.inputs.voxel_order = 'LAS'

    cam2trk_pico = pe.Node(
        interface=cam2trk.Camino2Trackvis(), name="cam2trk_pico")
    cam2trk_pico.inputs.min_length = 30
    cam2trk_pico.inputs.voxel_order = 'LAS'
    """
    Tracts can also be converted to VTK and OOGL formats, for use in programs such as GeomView and      Paraview, using the following two nodes.
    """

    # vtkstreamlines = pe.Node(interface=camino.VtkStreamlines(), name="vtkstreamlines")
    # procstreamlines = pe.Node(interface=camino.ProcStreamlines(), name="procstreamlines")
    # procstreamlines.inputs.outputtracts = 'oogl'
    """
    We can also produce a variety of scalar values from our fitted tensors. The following nodes generate        the fractional anisotropy and diffusivity trace maps and their associated headers.
    """

    fa = pe.Node(interface=camino.ComputeFractionalAnisotropy(), name='fa')
    # md = pe.Node(interface=camino.MD(),name='md')
    trace = pe.Node(interface=camino.ComputeTensorTrace(), name='trace')
    dteig = pe.Node(interface=camino.ComputeEigensystem(), name='dteig')

    analyzeheader_fa = pe.Node(
        interface=camino.AnalyzeHeader(), name="analyzeheader_fa")
    analyzeheader_fa.inputs.datatype = "double"
    analyzeheader_trace = analyzeheader_fa.clone('analyzeheader_trace')

    # analyzeheader_md = pe.Node(interface= camino.AnalyzeHeader(), name = "analyzeheader_md")
    # analyzeheader_md.inputs.datatype = "double"
    # analyzeheader_trace = analyzeheader_md.clone('analyzeheader_trace')

    fa2nii = pe.Node(interface=misc.CreateNifti(), name='fa2nii')
    trace2nii = fa2nii.clone("trace2nii")
    """
    Since we have now created all our nodes, we can now define our workflow and start making connections.
    """

    tractography = pe.Workflow(name='tractography')

    tractography.connect([(inputnode1, bet, [("dwi", "in_file")])])
    """
    File format conversion
    """

    tractography.connect([(inputnode1, image2voxel, [("dwi", "in_file")]),
                          (inputnode1, fsl2scheme, [("bvecs", "bvec_file"),
                                                    ("bvals", "bval_file")])])
    """
    Tensor fitting
    """

    tractography.connect([(image2voxel, dtifit, [['voxel_order', 'in_file']]),
                          (fsl2scheme, dtifit, [['scheme', 'scheme_file']])])
    """
    Workflow for applying DT streamline tractogpahy
    """

    tractography.connect([(bet, trackdt, [("mask_file", "seed_file")])])
    tractography.connect([(dtifit, trackdt, [("tensor_fitted", "in_file")])])
    """
    Workflow for applying PICo
    """

    tractography.connect([(bet, trackpico, [("mask_file", "seed_file")])])
    tractography.connect([(fsl2scheme, dtlutgen, [("scheme", "scheme_file")])])
    tractography.connect([(dtlutgen, picopdfs, [("dtLUT", "luts")])])
    tractography.connect([(dtifit, picopdfs, [("tensor_fitted", "in_file")])])
    tractography.connect([(picopdfs, trackpico, [("pdfs", "in_file")])])

    # Mean diffusivity still appears broken
    # tractography.connect([(dtifit, md,[("tensor_fitted","in_file")])])
    # tractography.connect([(md, analyzeheader_md,[("md","in_file")])])
    # tractography.connect([(inputnode, analyzeheader_md,[(('dwi', get_vox_dims), 'voxel_dims'),
    # (('dwi', get_data_dims), 'data_dims')])])
    # This line is commented out because the ProcStreamlines node keeps throwing memory errors
    # tractography.connect([(track, procstreamlines,[("tracked","in_file")])])
    """
    Connecting the Fractional Anisotropy and Trace nodes is simple, as they obtain their input from the
    tensor fitting.

    This is also where our voxel- and data-grabbing functions come in. We pass these functions, along       with the original DWI image from the input node, to the header-generating nodes. This ensures that      the files will be correct and readable.
    """

    tractography.connect([(dtifit, fa, [("tensor_fitted", "in_file")])])
    tractography.connect([(fa, analyzeheader_fa, [("fa", "in_file")])])
    tractography.connect([(inputnode1, analyzeheader_fa,
                           [(('dwi', get_vox_dims), 'voxel_dims'),
                            (('dwi', get_data_dims), 'data_dims')])])
    tractography.connect([(fa, fa2nii, [('fa', 'data_file')])])
    tractography.connect([(inputnode1, fa2nii, [(('dwi', get_affine),
                                                 'affine')])])
    tractography.connect([(analyzeheader_fa, fa2nii, [('header',
                                                       'header_file')])])

    tractography.connect([(dtifit, trace, [("tensor_fitted", "in_file")])])
    tractography.connect([(trace, analyzeheader_trace, [("trace",
                                                         "in_file")])])
    tractography.connect([(inputnode1, analyzeheader_trace,
                           [(('dwi', get_vox_dims), 'voxel_dims'),
                            (('dwi', get_data_dims), 'data_dims')])])
    tractography.connect([(trace, trace2nii, [('trace', 'data_file')])])
    tractography.connect([(inputnode1, trace2nii, [(('dwi', get_affine),
                                                    'affine')])])
    tractography.connect([(analyzeheader_trace, trace2nii, [('header',
                                                             'header_file')])])

    tractography.connect([(dtifit, dteig, [("tensor_fitted", "in_file")])])

    tractography.connect([(trackpico, cam2trk_pico, [('tracked', 'in_file')])])
    tractography.connect([(trackdt, cam2trk_dt, [('tracked', 'in_file')])])
    tractography.connect([(inputnode1, cam2trk_pico,
                           [(('dwi', get_vox_dims), 'voxel_dims'),
                            (('dwi', get_data_dims), 'data_dims')])])

    tractography.connect([(inputnode1, cam2trk_dt,
                           [(('dwi', get_vox_dims), 'voxel_dims'),
                            (('dwi', get_data_dims), 'data_dims')])])

    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=["dwi", "bvecs", "bvals"]),
        name="inputnode")

    outputnode = pe.Node(
        interface=util.IdentityInterface(
            fields=["fa", "trace", "tracts_pico", "tracts_dt", "tensors"]),
        name="outputnode")

    workflow = pe.Workflow(name=name)
    workflow.base_output_dir = name

    workflow.connect([(inputnode, tractography,
                       [("dwi", "inputnode1.dwi"),
                        ("bvals", "inputnode1.bvals"), ("bvecs",
                                                        "inputnode1.bvecs")])])

    workflow.connect([(tractography, outputnode,
                       [("cam2trk_dt.trackvis", "tracts_dt"),
                        ("cam2trk_pico.trackvis",
                         "tracts_pico"), ("fa2nii.nifti_file", "fa"),
                        ("trace2nii.nifti_file",
                         "trace"), ("dtifit.tensor_fitted", "tensors")])])

    return workflow



# -*- coding: utf-8 -*-
# coding: utf-8

from __future__ import absolute_import

from ....pipeline import engine as pe
from ....interfaces import utility as niu
from ....interfaces import fsl
from ....algorithms import misc

# backwards compatibility
from .epi import create_eddy_correct_pipeline


def transpose(samples_over_fibres):
    import numpy as np
    a = np.array(samples_over_fibres)
    return np.squeeze(a.T).tolist()


def create_bedpostx_pipeline(
        name='bedpostx',
        params={
            'n_fibres': 2,
            'fudge': 1,
            'burn_in': 1000,
            'n_jumps': 1250,
            'sample_every': 25,
            'model': 2,
            'cnlinear': True
        }):
    """
    Creates a pipeline that does the same as bedpostx script from FSL -
    calculates diffusion model parameters (distributions not MLE) voxelwise for
    the whole volume (by splitting it slicewise).

    Example
    -------

    >>> from nipype.workflows.dmri.fsl.dti import create_bedpostx_pipeline
    >>> params = dict(n_fibres = 2, fudge = 1, burn_in = 1000,
    ...               n_jumps = 1250, sample_every = 25)
    >>> bpwf = create_bedpostx_pipeline('nipype_bedpostx', params)
    >>> bpwf.inputs.inputnode.dwi = 'diffusion.nii'
    >>> bpwf.inputs.inputnode.mask = 'mask.nii'
    >>> bpwf.inputs.inputnode.bvecs = 'bvecs'
    >>> bpwf.inputs.inputnode.bvals = 'bvals'
    >>> bpwf.run() # doctest: +SKIP

    Inputs::

        inputnode.dwi
        inputnode.mask
        inputnode.bvecs
        inputnode.bvals

    Outputs::

        outputnode wraps all XFibres outputs

    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['dwi', 'mask', 'bvecs', 'bvals']),
        name='inputnode')

    slice_dwi = pe.Node(fsl.Split(dimension='z'), name='slice_dwi')
    slice_msk = pe.Node(fsl.Split(dimension='z'), name='slice_msk')
    mask_dwi = pe.MapNode(
        fsl.ImageMaths(op_string='-mas'),
        iterfield=['in_file', 'in_file2'],
        name='mask_dwi')

    xfib_if = fsl.XFibres(**params)
    xfibres = pe.MapNode(xfib_if, name='xfibres', iterfield=['dwi', 'mask'])

    make_dyads = pe.MapNode(
        fsl.MakeDyadicVectors(),
        name="make_dyads",
        iterfield=['theta_vol', 'phi_vol'])
    out_fields = [
        'dyads', 'dyads_disp', 'thsamples', 'phsamples', 'fsamples',
        'mean_thsamples', 'mean_phsamples', 'mean_fsamples'
    ]

    outputnode = pe.Node(
        niu.IdentityInterface(fields=out_fields), name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect(
        [(inputnode, slice_dwi, [('dwi', 'in_file')]), (inputnode, slice_msk,
                                                        [('mask', 'in_file')]),
         (slice_dwi, mask_dwi,
          [('out_files', 'in_file')]), (slice_msk, mask_dwi, [('out_files',
                                                               'in_file2')]),
         (slice_dwi, xfibres,
          [('out_files', 'dwi')]), (mask_dwi, xfibres, [('out_file', 'mask')]),
         (inputnode, xfibres, [('bvecs', 'bvecs'),
                               ('bvals', 'bvals')]), (inputnode, make_dyads,
                                                      [('mask', 'mask')])])

    mms = {}
    for k in ['thsamples', 'phsamples', 'fsamples']:
        mms[k] = merge_and_mean(k)
        wf.connect([(xfibres, mms[k], [(k, 'inputnode.in_files')]),
                    (mms[k], outputnode, [('outputnode.merged', k),
                                          ('outputnode.mean',
                                           'mean_%s' % k)])])

    # m_mdsamples = pe.Node(fsl.Merge(dimension="z"),
    #                       name="merge_mean_dsamples")
    wf.connect([
        (mms['thsamples'], make_dyads, [('outputnode.merged', 'theta_vol')]),
        (mms['phsamples'], make_dyads, [('outputnode.merged', 'phi_vol')]),
        # (xfibres, m_mdsamples,  [('mean_dsamples', 'in_files')]),
        (make_dyads, outputnode, [('dyads', 'dyads'), ('dispersion',
                                                       'dyads_disp')])
    ])
    return wf


def merge_and_mean(name='mm'):
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_files']), name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['merged', 'mean']), name='outputnode')
    merge = pe.MapNode(
        fsl.Merge(dimension='z'), name='Merge', iterfield=['in_files'])
    mean = pe.MapNode(
        fsl.ImageMaths(op_string='-Tmean'), name='Mean', iterfield=['in_file'])

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, merge, [(('in_files', transpose), 'in_files')]),
                (merge, mean, [('merged_file', 'in_file')]),
                (merge, outputnode,
                 [('merged_file', 'merged')]), (mean, outputnode, [('out_file',
                                                                    'mean')])])
    return wf


def bedpostx_parallel(
        name='bedpostx_parallel',
        compute_all_outputs=True,
        params={
            'n_fibres': 2,
            'fudge': 1,
            'burn_in': 1000,
            'n_jumps': 1250,
            'sample_every': 25,
            'model': 1,
            'cnlinear': True
        }):
    """
    Does the same as :func:`.create_bedpostx_pipeline` by splitting
    the input dMRI in small ROIs that are better suited for parallel
    processing).

    Example
    -------

    >>> from nipype.workflows.dmri.fsl.dti import bedpostx_parallel
    >>> params = dict(n_fibres = 2, fudge = 1, burn_in = 1000,
    ...               n_jumps = 1250, sample_every = 25)
    >>> bpwf = bedpostx_parallel('nipype_bedpostx_parallel', params=params)
    >>> bpwf.inputs.inputnode.dwi = 'diffusion.nii'
    >>> bpwf.inputs.inputnode.mask = 'mask.nii'
    >>> bpwf.inputs.inputnode.bvecs = 'bvecs'
    >>> bpwf.inputs.inputnode.bvals = 'bvals'
    >>> bpwf.run(plugin='CondorDAGMan') # doctest: +SKIP

    Inputs::

        inputnode.dwi
        inputnode.mask
        inputnode.bvecs
        inputnode.bvals

    Outputs::

        outputnode wraps all XFibres outputs

    """

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['dwi', 'mask', 'bvecs', 'bvals']),
        name='inputnode')
    slice_dwi = pe.Node(misc.SplitROIs(roi_size=(5, 5, 1)), name='slice_dwi')
    if params is not None:
        xfib_if = fsl.XFibres5(**params)
    else:
        xfib_if = fsl.XFibres5()
    xfibres = pe.MapNode(xfib_if, name='xfibres', iterfield=['dwi', 'mask'])

    mrg_dyads = pe.MapNode(
        misc.MergeROIs(), name='Merge_dyads', iterfield=['in_files'])
    mrg_fsamp = pe.MapNode(
        misc.MergeROIs(), name='Merge_mean_fsamples', iterfield=['in_files'])
    out_fields = ['dyads', 'fsamples']

    if compute_all_outputs:
        out_fields += [
            'dyads_disp', 'thsamples', 'phsamples', 'mean_fsamples',
            'mean_thsamples', 'mean_phsamples', 'merged_fsamples',
            'merged_thsamples', 'merged_phsamples'
        ]

    outputnode = pe.Node(
        niu.IdentityInterface(fields=out_fields), name='outputnode')

    wf = pe.Workflow(name=name)
    wf.connect(
        [(inputnode, slice_dwi, [('dwi', 'in_file'), ('mask', 'in_mask')]),
         (slice_dwi, xfibres, [('out_files', 'dwi'), ('out_masks', 'mask')]),
         (inputnode, xfibres,
          [('bvecs', 'bvecs'), ('bvals', 'bvals')]), (inputnode, mrg_dyads, [
              ('mask', 'in_reference')
          ]), (xfibres, mrg_dyads,
               [(('dyads', transpose), 'in_files')]), (slice_dwi, mrg_dyads, [
                   ('out_index', 'in_index')
               ]), (inputnode, mrg_fsamp,
                    [('mask', 'in_reference')]), (xfibres, mrg_fsamp, [
                        (('mean_fsamples', transpose), 'in_files')
                    ]), (slice_dwi, mrg_fsamp, [('out_index', 'in_index')]),
         (mrg_dyads, outputnode,
          [('merged_file', 'dyads')]), (mrg_fsamp, outputnode,
                                        [('merged_file', 'fsamples')])])

    if compute_all_outputs:
        make_dyads = pe.MapNode(
            fsl.MakeDyadicVectors(),
            name="Make_dyads",
            iterfield=['theta_vol', 'phi_vol'])

        wf.connect([(inputnode, make_dyads, [('mask', 'mask')])])
        mms = {}
        for k in ['thsamples', 'phsamples', 'fsamples']:
            mms[k] = merge_and_mean_parallel(k)
            wf.connect(
                [(slice_dwi, mms[k], [('out_index', 'inputnode.in_index')]),
                 (inputnode, mms[k], [('mask', 'inputnode.in_reference')]),
                 (xfibres, mms[k], [(k, 'inputnode.in_files')]),
                 (mms[k], outputnode, [('outputnode.merged', 'merged_%s' % k),
                                       ('outputnode.mean', 'mean_%s' % k)])])

        # m_mdsamples = pe.Node(fsl.Merge(dimension="z"),
        #                       name="merge_mean_dsamples")
        wf.connect([
            (mms['thsamples'], make_dyads, [('outputnode.merged',
                                             'theta_vol')]),
            (mms['phsamples'], make_dyads, [('outputnode.merged', 'phi_vol')]),
            # (xfibres, m_mdsamples,  [('mean_dsamples', 'in_files')]),
            (make_dyads, outputnode, [('dispersion', 'dyads_disp')])
        ])

    return wf


def merge_and_mean_parallel(name='mm'):
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_files', 'in_reference', 'in_index']),
        name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['merged', 'mean']), name='outputnode')
    merge = pe.MapNode(misc.MergeROIs(), name='Merge', iterfield=['in_files'])
    mean = pe.MapNode(
        fsl.ImageMaths(op_string='-Tmean'), name='Mean', iterfield=['in_file'])

    wf = pe.Workflow(name=name)
    wf.connect([(inputnode, merge,
                 [(('in_files', transpose), 'in_files'),
                  ('in_reference', 'in_reference'), ('in_index', 'in_index')]),
                (merge, mean, [('merged_file', 'in_file')]),
                (merge, outputnode,
                 [('merged_file', 'merged')]), (mean, outputnode, [('out_file',
                                                                    'mean')])])
    return wf



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import os
from warnings import warn

from ....pipeline import engine as pe
from ....interfaces import utility as util
from ....interfaces import fsl as fsl


def tbss1_op_string(in_files):
    import nibabel as nb
    from nipype.utils import NUMPY_MMAP
    op_strings = []
    for infile in in_files:
        img = nb.load(infile, mmap=NUMPY_MMAP)
        dimtup = tuple(d - 2 for d in img.shape)
        dimtup = dimtup[0:3]
        op_str = '-min 1 -ero -roi 1 %d 1 %d 1 %d 0 1' % dimtup
        op_strings.append(op_str)
    return op_strings


def create_tbss_1_preproc(name='tbss_1_preproc'):
    """Preprocess FA data for TBSS: erodes a little and zero end slicers and
    creates masks(for use in FLIRT & FNIRT from FSL).
    A pipeline that does the same as tbss_1_preproc script in FSL

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss1 = tbss.create_tbss_1_preproc()
    >>> tbss1.inputs.inputnode.fa_list = ['s1_FA.nii', 's2_FA.nii', 's3_FA.nii']

    Inputs::

        inputnode.fa_list

    Outputs::

        outputnode.fa_list
        outputnode.mask_list
        outputnode.slices

    """

    # Define the inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=["fa_list"]), name="inputnode")

    # Prep the FA images
    prepfa = pe.MapNode(
        fsl.ImageMaths(suffix="_prep"),
        name="prepfa",
        iterfield=['in_file', 'op_string'])

    # Slicer
    slicer = pe.MapNode(
        fsl.Slicer(all_axial=True, image_width=1280),
        name='slicer',
        iterfield=['in_file'])

    # Create a mask
    getmask1 = pe.MapNode(
        fsl.ImageMaths(op_string="-bin", suffix="_mask"),
        name="getmask1",
        iterfield=['in_file'])
    getmask2 = pe.MapNode(
        fsl.MultiImageMaths(op_string="-dilD -dilD -sub 1 -abs -add %s"),
        name="getmask2",
        iterfield=['in_file', 'operand_files'])

    # $FSLDIR/bin/fslmaths FA/${f}_FA_mask -dilD -dilD -sub 1 -abs -add FA/${f}_FA_mask FA/${f}_FA_mask -odt char
    # Define the tbss1 workflow
    tbss1 = pe.Workflow(name=name)
    tbss1.connect([
        (inputnode, prepfa, [("fa_list", "in_file")]),
        (inputnode, prepfa, [(("fa_list", tbss1_op_string), "op_string")]),
        (prepfa, getmask1, [("out_file", "in_file")]),
        (getmask1, getmask2, [("out_file", "in_file"), ("out_file",
                                                        "operand_files")]),
        (prepfa, slicer, [('out_file', 'in_file')]),
    ])

    # Define the outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(
            fields=["fa_list", "mask_list", "slices"]),
        name="outputnode")
    tbss1.connect([(prepfa, outputnode, [("out_file", "fa_list")]),
                   (getmask2, outputnode, [("out_file", "mask_list")]),
                   (slicer, outputnode, [('out_file', 'slices')])])
    return tbss1


def create_tbss_2_reg(name="tbss_2_reg"):
    """TBSS nonlinear registration:
    A pipeline that does the same as 'tbss_2_reg -t' script in FSL. '-n' option
    is not supported at the moment.

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss2 = create_tbss_2_reg(name="tbss2")
    >>> tbss2.inputs.inputnode.target = fsl.Info.standard_image("FMRIB58_FA_1mm.nii.gz")  # doctest: +SKIP
    >>> tbss2.inputs.inputnode.fa_list = ['s1_FA.nii', 's2_FA.nii', 's3_FA.nii']
    >>> tbss2.inputs.inputnode.mask_list = ['s1_mask.nii', 's2_mask.nii', 's3_mask.nii']

    Inputs::

        inputnode.fa_list
        inputnode.mask_list
        inputnode.target

    Outputs::

        outputnode.field_list

    """

    # Define the inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(
            fields=["fa_list", "mask_list", "target"]),
        name="inputnode")

    # Flirt the FA image to the target
    flirt = pe.MapNode(
        interface=fsl.FLIRT(dof=12),
        iterfield=['in_file', 'in_weight'],
        name="flirt")

    fnirt = pe.MapNode(
        interface=fsl.FNIRT(fieldcoeff_file=True),
        iterfield=['in_file', 'inmask_file', 'affine_file'],
        name="fnirt")
    # Fnirt the FA image to the target
    if fsl.no_fsl():
        warn('NO FSL found')
    else:
        config_file = os.path.join(os.environ["FSLDIR"],
                                   "etc/flirtsch/FA_2_FMRIB58_1mm.cnf")
        fnirt.inputs.config_file = config_file

    # Define the registration workflow
    tbss2 = pe.Workflow(name=name)

    # Connect up the registration workflow
    tbss2.connect([
        (inputnode, flirt, [("fa_list", "in_file"), ("target", "reference"),
                            ("mask_list", "in_weight")]),
        (inputnode, fnirt, [("fa_list", "in_file"),
                            ("mask_list", "inmask_file"), ("target",
                                                           "ref_file")]),
        (flirt, fnirt, [("out_matrix_file", "affine_file")]),
    ])

    # Define the outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=['field_list']),
        name="outputnode")

    tbss2.connect([(fnirt, outputnode, [('fieldcoeff_file', 'field_list')])])
    return tbss2


def create_tbss_3_postreg(name='tbss_3_postreg', estimate_skeleton=True):
    """Post-registration processing: derive mean_FA and mean_FA_skeleton from
    mean of all subjects in study. Target is assumed to be FMRIB58_FA_1mm.
    A pipeline that does the same as 'tbss_3_postreg -S' script from FSL
    Setting 'estimate_skeleton to False will use precomputed FMRIB58_FA-skeleton_1mm
    skeleton (same as 'tbss_3_postreg -T').

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss3 = tbss.create_tbss_3_postreg()
    >>> tbss3.inputs.inputnode.fa_list = ['s1_wrapped_FA.nii', 's2_wrapped_FA.nii', 's3_wrapped_FA.nii']

    Inputs::

        inputnode.field_list
        inputnode.fa_list

    Outputs::

        outputnode.groupmask
        outputnode.skeleton_file
        outputnode.meanfa_file
        outputnode.mergefa_file

    """

    # Create the inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=['field_list', 'fa_list']),
        name='inputnode')

    # Apply the warpfield to the masked FA image
    applywarp = pe.MapNode(
        interface=fsl.ApplyWarp(),
        iterfield=['in_file', 'field_file'],
        name="applywarp")
    if fsl.no_fsl():
        warn('NO FSL found')
    else:
        applywarp.inputs.ref_file = fsl.Info.standard_image(
            "FMRIB58_FA_1mm.nii.gz")

    # Merge the FA files into a 4D file
    mergefa = pe.Node(fsl.Merge(dimension="t"), name="mergefa")

    # Get a group mask
    groupmask = pe.Node(
        fsl.ImageMaths(
            op_string="-max 0 -Tmin -bin",
            out_data_type="char",
            suffix="_mask"),
        name="groupmask")

    maskgroup = pe.Node(
        fsl.ImageMaths(op_string="-mas", suffix="_masked"), name="maskgroup")

    tbss3 = pe.Workflow(name=name)
    tbss3.connect([
        (inputnode, applywarp, [("fa_list", "in_file"), ("field_list",
                                                         "field_file")]),
        (applywarp, mergefa, [("out_file", "in_files")]),
        (mergefa, groupmask, [("merged_file", "in_file")]),
        (mergefa, maskgroup, [("merged_file", "in_file")]),
        (groupmask, maskgroup, [("out_file", "in_file2")]),
    ])

    # Create outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'groupmask', 'skeleton_file', 'meanfa_file', 'mergefa_file'
        ]),
        name='outputnode')

    if estimate_skeleton:
        # Take the mean over the fourth dimension
        meanfa = pe.Node(
            fsl.ImageMaths(op_string="-Tmean", suffix="_mean"), name="meanfa")

        # Use the mean FA volume to generate a tract skeleton
        makeskeleton = pe.Node(
            fsl.TractSkeleton(skeleton_file=True), name="makeskeleton")
        tbss3.connect(
            [(maskgroup, meanfa, [("out_file", "in_file")]),
             (meanfa, makeskeleton,
              [("out_file", "in_file")]), (groupmask, outputnode,
                                           [('out_file', 'groupmask')]),
             (makeskeleton, outputnode,
              [('skeleton_file', 'skeleton_file')]), (meanfa, outputnode, [
                  ('out_file', 'meanfa_file')
              ]), (maskgroup, outputnode, [('out_file', 'mergefa_file')])])
    else:
        # $FSLDIR/bin/fslmaths $FSLDIR/data/standard/FMRIB58_FA_1mm -mas mean_FA_mask mean_FA
        maskstd = pe.Node(
            fsl.ImageMaths(op_string="-mas", suffix="_masked"), name="maskstd")
        if fsl.no_fsl():
            warn('NO FSL found')
        else:
            maskstd.inputs.in_file = fsl.Info.standard_image(
                "FMRIB58_FA_1mm.nii.gz")

        # $FSLDIR/bin/fslmaths mean_FA -bin mean_FA_mask
        binmaskstd = pe.Node(
            fsl.ImageMaths(op_string="-bin"), name="binmaskstd")

        # $FSLDIR/bin/fslmaths all_FA -mas mean_FA_mask all_FA
        maskgroup2 = pe.Node(
            fsl.ImageMaths(op_string="-mas", suffix="_masked"),
            name="maskgroup2")

        tbss3.connect([(groupmask, maskstd, [("out_file", "in_file2")]),
                       (maskstd, binmaskstd, [("out_file", "in_file")]),
                       (maskgroup, maskgroup2, [("out_file", "in_file")]),
                       (binmaskstd, maskgroup2, [("out_file", "in_file2")])])

        if fsl.no_fsl():
            warn('NO FSL found')
        else:
            outputnode.inputs.skeleton_file = fsl.Info.standard_image(
                "FMRIB58_FA-skeleton_1mm.nii.gz")
        tbss3.connect([(binmaskstd, outputnode, [('out_file', 'groupmask')]),
                       (maskstd, outputnode, [('out_file', 'meanfa_file')]),
                       (maskgroup2, outputnode, [('out_file',
                                                  'mergefa_file')])])
    return tbss3


def tbss4_op_string(skeleton_thresh):
    op_string = "-thr %.1f -bin" % skeleton_thresh
    return op_string


def create_tbss_4_prestats(name='tbss_4_prestats'):
    """Post-registration processing:Creating skeleton mask using a threshold
     projecting all FA data onto skeleton.
    A pipeline that does the same as tbss_4_prestats script from FSL

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss4 = tbss.create_tbss_4_prestats(name='tbss4')
    >>> tbss4.inputs.inputnode.skeleton_thresh = 0.2

    Inputs::

        inputnode.skeleton_thresh
        inputnode.groupmask
        inputnode.skeleton_file
        inputnode.meanfa_file
        inputnode.mergefa_file

    Outputs::

        outputnode.all_FA_skeletonised
        outputnode.mean_FA_skeleton_mask
        outputnode.distance_map
        outputnode.skeleton_file

    """
    # Create inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'groupmask', 'skeleton_file', 'meanfa_file', 'mergefa_file',
            'skeleton_thresh'
        ]),
        name='inputnode')

    # Mask the skeleton at the threshold
    skeletonmask = pe.Node(fsl.ImageMaths(suffix="_mask"), name="skeletonmask")

    # Invert the brainmask then add in the tract skeleton
    invertmask = pe.Node(
        fsl.ImageMaths(suffix="_inv", op_string="-mul -1 -add 1 -add"),
        name="invertmask")

    # Generate a distance map with the tract skeleton
    distancemap = pe.Node(fsl.DistanceMap(), name="distancemap")

    # Project the FA values onto the skeleton
    projectfa = pe.Node(
        fsl.TractSkeleton(
            project_data=True, skeleton_file=True, use_cingulum_mask=True),
        name="projectfa")

    # Create tbss4 workflow
    tbss4 = pe.Workflow(name=name)
    tbss4.connect([
        (inputnode, invertmask, [("groupmask", "in_file")]),
        (inputnode, skeletonmask, [("skeleton_file", "in_file"),
                                   (('skeleton_thresh', tbss4_op_string),
                                    'op_string')]),
        (inputnode, projectfa, [('skeleton_thresh', 'threshold'),
                                ("meanfa_file", "in_file"), ("mergefa_file",
                                                             "data_file")]),
        (skeletonmask, invertmask, [("out_file", "in_file2")]),
        (invertmask, distancemap, [("out_file", "in_file")]),
        (distancemap, projectfa, [("distance_map", "distance_map")]),
    ])

    # Create the outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'projectedfa_file', 'skeleton_mask', 'distance_map',
            'skeleton_file'
        ]),
        name='outputnode')

    tbss4.connect(
        [(projectfa, outputnode, [('projected_data', 'projectedfa_file'),
                                  ('skeleton_file', 'skeleton_file')]),
         (distancemap, outputnode, [('distance_map', 'distance_map')]),
         (skeletonmask, outputnode, [('out_file', 'skeleton_mask')])])

    return tbss4


def create_tbss_all(name='tbss_all', estimate_skeleton=True):
    """Create a pipeline that combines create_tbss_* pipelines

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss_wf = tbss.create_tbss_all('tbss', estimate_skeleton=True)
    >>> tbss_wf.inputs.inputnode.skeleton_thresh = 0.2
    >>> tbss_wf.inputs.inputnode.fa_list = ['s1_wrapped_FA.nii', 's2_wrapped_FA.nii', 's3_wrapped_FA.nii']

    >>> tbss_wf = tbss.create_tbss_all('tbss', estimate_skeleton=False)
    >>> tbss_wf.inputs.inputnode.skeleton_thresh = 0.2
    >>> tbss_wf.inputs.inputnode.fa_list = ['s1_wrapped_FA.nii', 's2_wrapped_FA.nii', 's3_wrapped_FA.nii']


    Inputs::

        inputnode.fa_list
        inputnode.skeleton_thresh

    Outputs::

        outputnode.meanfa_file
        outputnode.projectedfa_file
        outputnode.skeleton_file
        outputnode.skeleton_mask

    """

    # Define the inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(
            fields=['fa_list', 'skeleton_thresh']),
        name='inputnode')

    tbss1 = create_tbss_1_preproc(name='tbss1')
    tbss2 = create_tbss_2_reg(name='tbss2')
    if fsl.no_fsl():
        warn('NO FSL found')
    else:
        tbss2.inputs.inputnode.target = fsl.Info.standard_image(
            "FMRIB58_FA_1mm.nii.gz")
    tbss3 = create_tbss_3_postreg(
        name='tbss3', estimate_skeleton=estimate_skeleton)
    tbss4 = create_tbss_4_prestats(name='tbss4')

    tbss_all = pe.Workflow(name=name)
    tbss_all.connect(
        [(inputnode, tbss1, [('fa_list', 'inputnode.fa_list')]),
         (inputnode, tbss4,
          [('skeleton_thresh', 'inputnode.skeleton_thresh')]), (tbss1, tbss2, [
              ('outputnode.fa_list', 'inputnode.fa_list'),
              ('outputnode.mask_list', 'inputnode.mask_list')
          ]), (tbss1, tbss3, [('outputnode.fa_list', 'inputnode.fa_list')]),
         (tbss2, tbss3, [('outputnode.field_list', 'inputnode.field_list')]),
         (tbss3, tbss4, [('outputnode.groupmask',
                          'inputnode.groupmask'), ('outputnode.skeleton_file',
                                                   'inputnode.skeleton_file'),
                         ('outputnode.meanfa_file', 'inputnode.meanfa_file'),
                         ('outputnode.mergefa_file',
                          'inputnode.mergefa_file')])])

    # Define the outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'groupmask', 'skeleton_file3', 'meanfa_file', 'mergefa_file',
            'projectedfa_file', 'skeleton_file4', 'skeleton_mask',
            'distance_map'
        ]),
        name='outputnode')
    outputall_node = pe.Node(
        interface=util.IdentityInterface(fields=[
            'fa_list1', 'mask_list1', 'field_list2', 'groupmask3',
            'skeleton_file3', 'meanfa_file3', 'mergefa_file3',
            'projectedfa_file4', 'skeleton_mask4', 'distance_map4'
        ]),
        name='outputall_node')

    tbss_all.connect([
        (tbss3, outputnode, [
            ('outputnode.meanfa_file', 'meanfa_file'),
            ('outputnode.mergefa_file', 'mergefa_file'),
            ('outputnode.groupmask', 'groupmask'),
            ('outputnode.skeleton_file', 'skeleton_file3'),
        ]),
        (tbss4, outputnode, [
            ('outputnode.projectedfa_file', 'projectedfa_file'),
            ('outputnode.skeleton_file', 'skeleton_file4'),
            ('outputnode.skeleton_mask', 'skeleton_mask'),
            ('outputnode.distance_map', 'distance_map'),
        ]),
        (tbss1, outputall_node, [
            ('outputnode.fa_list', 'fa_list1'),
            ('outputnode.mask_list', 'mask_list1'),
        ]),
        (tbss2, outputall_node, [
            ('outputnode.field_list', 'field_list2'),
        ]),
        (tbss3, outputall_node, [
            ('outputnode.meanfa_file', 'meanfa_file3'),
            ('outputnode.mergefa_file', 'mergefa_file3'),
            ('outputnode.groupmask', 'groupmask3'),
            ('outputnode.skeleton_file', 'skeleton_file3'),
        ]),
        (tbss4, outputall_node, [
            ('outputnode.projectedfa_file', 'projectedfa_file4'),
            ('outputnode.skeleton_mask', 'skeleton_mask4'),
            ('outputnode.distance_map', 'distance_map4'),
        ]),
    ])
    return tbss_all


def create_tbss_non_FA(name='tbss_non_FA'):
    """
    A pipeline that implement tbss_non_FA in FSL

    Example
    -------

    >>> from nipype.workflows.dmri.fsl import tbss
    >>> tbss_MD = tbss.create_tbss_non_FA()
    >>> tbss_MD.inputs.inputnode.file_list = []
    >>> tbss_MD.inputs.inputnode.field_list = []
    >>> tbss_MD.inputs.inputnode.skeleton_thresh = 0.2
    >>> tbss_MD.inputs.inputnode.groupmask = './xxx'
    >>> tbss_MD.inputs.inputnode.meanfa_file = './xxx'
    >>> tbss_MD.inputs.inputnode.distance_map = []
    >>> tbss_MD.inputs.inputnode.all_FA_file = './xxx'

    Inputs::

        inputnode.file_list
        inputnode.field_list
        inputnode.skeleton_thresh
        inputnode.groupmask
        inputnode.meanfa_file
        inputnode.distance_map
        inputnode.all_FA_file

    Outputs::

        outputnode.projected_nonFA_file

    """

    # Define the inputnode
    inputnode = pe.Node(
        interface=util.IdentityInterface(fields=[
            'file_list', 'field_list', 'skeleton_thresh', 'groupmask',
            'meanfa_file', 'distance_map', 'all_FA_file'
        ]),
        name='inputnode')

    # Apply the warpfield to the non FA image
    applywarp = pe.MapNode(
        interface=fsl.ApplyWarp(),
        iterfield=['in_file', 'field_file'],
        name="applywarp")
    if fsl.no_fsl():
        warn('NO FSL found')
    else:
        applywarp.inputs.ref_file = fsl.Info.standard_image(
            "FMRIB58_FA_1mm.nii.gz")
    # Merge the non FA files into a 4D file
    merge = pe.Node(fsl.Merge(dimension="t"), name="merge")
    # merged_file="all_FA.nii.gz"
    maskgroup = pe.Node(
        fsl.ImageMaths(op_string="-mas", suffix="_masked"), name="maskgroup")
    projectfa = pe.Node(
        fsl.TractSkeleton(
            project_data=True,
            # projected_data = 'test.nii.gz',
            use_cingulum_mask=True),
        name="projectfa")

    tbss_non_FA = pe.Workflow(name=name)
    tbss_non_FA.connect([
        (inputnode, applywarp, [
            ('file_list', 'in_file'),
            ('field_list', 'field_file'),
        ]),
        (applywarp, merge, [("out_file", "in_files")]),
        (merge, maskgroup, [("merged_file", "in_file")]),
        (inputnode, maskgroup, [('groupmask', 'in_file2')]),
        (maskgroup, projectfa, [('out_file', 'alt_data_file')]),
        (inputnode, projectfa,
         [('skeleton_thresh', 'threshold'), ("meanfa_file", "in_file"),
          ("distance_map", "distance_map"), ("all_FA_file", 'data_file')]),
    ])

    # Define the outputnode
    outputnode = pe.Node(
        interface=util.IdentityInterface(fields=['projected_nonFA_file']),
        name='outputnode')
    tbss_non_FA.connect([
        (projectfa, outputnode, [
            ('projected_data', 'projected_nonFA_file'),
        ]),
    ])
    return tbss_non_FA
